import logging
import sys
import timeit

from pymemtrace import trace_malloc

def create_string(l: int) -> str:
    return ' ' * l


COUNT = 16


def test_under_512(list_of_strings):
    # print(f'test_under_512 count={COUNT}')
    for i in range(COUNT):
        list_of_strings.append(create_string(256))
    # while len(list_of_strings):
    #     list_of_strings.pop()


def test_over_512(list_of_strings):
    # print(f'test_over_512 count={COUNT}')
    for i in range(COUNT):
        list_of_strings.append(create_string(1024))
    # while len(list_of_strings):
    #     list_of_strings.pop()


def test_well_over_512(list_of_strings):
    # print(f'test_well_over_512 count={COUNT}')
    for i in range(COUNT):
        list_of_strings.append(create_string(1024**2))
    # while len(list_of_strings):
    #     list_of_strings.pop()


def example_trace_malloc_for_documentation(list_of_strings):
    """An example of using the trace_malloc.trace_malloc_report decorator for logging memory usage.
    Typical output::

        example_trace_malloc_for_documentation()
        pymemtrace/pymemtrace/examples/example_trace_malloc.py:0: size=8194 KiB (+8193 KiB), count=16 (+10), average=512 KiB
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/tracemalloc.py:0: size=6720 B (+552 B), count=43 (+11), average=156 B
        pymemtrace/pymemtrace/trace_malloc.py:0: size=3076 B (-468 B), count=10 (-1), average=308 B
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/logging/__init__.py:0: size=16.3 KiB (-176 B), count=49 (-3), average=340 B
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/abc.py:0: size=3169 B (+0 B), count=30 (+0), average=106 B
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/posixpath.py:0: size=480 B (+0 B), count=1 (+0), average=480 B
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/threading.py:0: size=168 B (+0 B), count=2 (+0), average=84 B
        /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/_weakrefset.py:0: size=72 B (+0 B), count=1 (+0), average=72 B
    """
    print(f'example_trace_malloc_for_documentation()')
    with trace_malloc.TraceMalloc('filename') as tm:
        for i in range(8):
            list_of_strings.append(create_string(1024**2))
            # list_of_strings.append(create_string(128))
    print(f' tm.memory_start={tm.memory_start}')
    print(f'tm.memory_finish={tm.memory_finish}')
    print(f'         tm.diff={tm.diff}')
    for stat in tm.statistics:
        print(stat)
    print()
    for stat in tm.net_statistics():
        print(stat)


@trace_malloc.trace_malloc_log(logging.INFO)
def example_decorator_for_documentation(list_of_strings):
    """An example of using the trace_malloc.trace_malloc_report decorator for logging memory usage.
    Typical output::

        2020-11-15 18:13:06,000 -          trace_malloc.py#82   -  9689 - (MainThread) - INFO     - TraceMalloc memory delta: 8,389,548
    """
    print(f'example_decorator_for_documentation()')
    for i in range(8):
        list_of_strings.append(create_string(1024**2))


def example():
    for function in (test_under_512, test_over_512, test_well_over_512):
        print(f'Function: {function}')
        list_of_strings = []
        # with trace_malloc.TraceMalloc('filename') as tm:
        with trace_malloc.TraceMalloc('lineno') as tm:
            function(list_of_strings)
        print(f'tm.memory_start={tm.memory_start}')
        print(f'tm.memory_finish={tm.memory_finish}')
        for stat in tm.net_statistics():
            print(stat)
        print()


def example_timeit_under_512():
    list_of_strings = []
    test_under_512(list_of_strings)


def example_timeit_under_512_with_trace_malloc(key_type):
    list_of_strings = []
    with trace_malloc.TraceMalloc(key_type) as tm:
        test_under_512(list_of_strings)


def example_timeit_over_512():
    list_of_strings = []
    test_over_512(list_of_strings)


def example_timeit_over_512_with_trace_malloc(key_type):
    list_of_strings = []
    with trace_malloc.TraceMalloc(key_type) as tm:
        test_over_512(list_of_strings)


def example_timeit_well_over_512():
    list_of_strings = []
    test_well_over_512(list_of_strings)


def example_timeit_well_over_512_with_trace_malloc(key_type):
    list_of_strings = []
    with trace_malloc.TraceMalloc(key_type) as tm:
        test_well_over_512(list_of_strings)


def run_timeit():
    NUMBER = 10_000
    REPEAT = 5
    CONVERT = 1_000_000
    print(f'number={NUMBER:,d} repeat={REPEAT:,d} convert={CONVERT:,d}')
    for function in (
            'example_timeit_under_512',
            'example_timeit_over_512',
            'example_timeit_well_over_512',
    ):
        # t = timeit.timeit(f"{function}()", setup=f"from __main__ import {function}", number=NUMBER) / NUMBER
        # print(f'{function:60}: {t:9.9f}')
        times = timeit.repeat(f"{function}()", setup=f"from __main__ import {function}", number=NUMBER, repeat=REPEAT)
        times = [CONVERT * t / NUMBER for t in times]
        result = [f'{v:9.3f}' for v in times]
        times_mean = sum(times) / REPEAT
        print(
            f'{function:60}:'
            f' {", ".join(result)}'
            f' mean={times_mean:9.3f}'
            f' min={min(times):9.3f}'
            f' max={max(times):9.3f}'
            f' span={max(times) - min(times):9.3f}'
        )
        for key_type in ('filename', 'lineno', 'traceback'):
            # With _with_trace_malloc
            times_with_trace_malloc = timeit.repeat(f"{function}_with_trace_malloc('{key_type}')", setup=f"from __main__ import {function}_with_trace_malloc", number=NUMBER, repeat=REPEAT)
            times_with_trace_malloc = [CONVERT * t / NUMBER for t in times_with_trace_malloc]
            result = [f'{v:9.3f}' for v in times_with_trace_malloc]
            times_mean_with_trace_malloc = sum(times_with_trace_malloc) / REPEAT
            function_str = function + f"_with_trace_malloc('{key_type}')"
            print(
                f'{function_str:60}:'
                f' {", ".join(result)}'
                f' mean={times_mean_with_trace_malloc:9.3f}'
                f' min={min(times_with_trace_malloc):9.3f}'
                f' max={max(times_with_trace_malloc):9.3f}'
                f' span={max(times_with_trace_malloc) - min(times_with_trace_malloc):9.3f}'
                f' x{times_mean_with_trace_malloc / times_mean:>8.3f}'
            )


def main() -> int:
    logging.basicConfig(
        level=logging.INFO,
        # format='%(asctime)s - %(filename)24s#%(lineno)-4d - %(funcName)24s - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
        format='%(asctime)s - %(filename)24s#%(lineno)-4d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
        stream=sys.stdout
    )
    print()
    example_decorator_for_documentation([])
    print()
    example_trace_malloc_for_documentation([])
    # print()
    # example()
    return 0


if __name__ == '__main__':
    sys.exit(main())




"""
Example of using process that logs process data to the current log.
"""
import logging
import random
import sys
import time

from pymemtrace import process

logger = logging.getLogger(__file__)


def main() -> int:
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(filename)s#%(lineno)d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
    )
    logger.info('Demonstration of logging a process')
    # Log process data to the log file every 0.5 seconds.
    with process.log_process(interval=0.5, log_level=logger.getEffectiveLevel()):
        for i in range(8):
            size = random.randint(128, 128 + 256) * 1024 ** 2
            # Add a message to report in the next process write.
            process.add_message_to_queue(f'String of {size:,d} bytes')
            s = ' ' * size
            time.sleep(0.75 + random.random())
            del s
            time.sleep(0.25 + random.random() / 2)
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Taken from the excellent blog:
https://eli.thegreenplace.net/2015/redirecting-all-kinds-of-stdout-in-python/

Changes:

* Minor edits.
* Move tfile = tempfile.TemporaryFile(mode='w+b') outside try block.
* Duplicate for stderr

TODO: Unite duplicate code.
"""
from contextlib import contextmanager
import ctypes
import io
import os
import sys
import tempfile

import typing

libc = ctypes.CDLL(None)
if sys.platform == 'darwin':
    c_stdout = ctypes.c_void_p.in_dll(libc, '__stdoutp')
    c_stderr = ctypes.c_void_p.in_dll(libc, '__stderrp')
else:
    c_stdout = ctypes.c_void_p.in_dll(libc, 'stdout')
    c_stderr = ctypes.c_void_p.in_dll(libc, 'stderr')


@contextmanager
def stdout_redirector(stream: typing.BinaryIO):
    """A context manager that redirects Python stdout and C stdout to the given binary I/O stream."""
    def _redirect_stdout(to_fd):
        """Redirect stdout to the given file descriptor."""
        # Flush the C-level buffer stdout
        libc.fflush(c_stdout)
        # Flush and close sys.stdout - also closes the file descriptor (fd)
        sys.stdout.close()
        # Make original_stdout_fd point to the same file as to_fd
        os.dup2(to_fd, original_stdout_fd)
        # Create a new sys.stdout that points to the redirected fd
        sys.stdout = io.TextIOWrapper(os.fdopen(original_stdout_fd, 'wb'))

    # The original fd stdout points to. Usually 1 on POSIX systems.
    original_stdout_fd = sys.stdout.fileno()
    # Save a copy of the original stdout fd in saved_stdout_fd
    saved_stdout_fd = os.dup(original_stdout_fd)
    # Create a temporary file and redirect stdout to it
    tfile = tempfile.TemporaryFile(mode='w+b')
    try:
        _redirect_stdout(tfile.fileno())
        # Yield to caller, then redirect stdout back to the saved fd
        yield
        _redirect_stdout(saved_stdout_fd)
        # Copy contents of temporary file to the given stream
        tfile.flush()
        tfile.seek(0, io.SEEK_SET)
        stream.write(tfile.read())
    finally:
        tfile.close()
        os.close(saved_stdout_fd)


@contextmanager
def stderr_redirector(stream: typing.BinaryIO):
    """A context manager that redirects Python stderr and C stderr to the given binary I/O stream."""
    def _redirect_stderr(to_fd):
        """Redirect stderr to the given file descriptor."""
        # Flush the C-level buffer stderr
        libc.fflush(c_stderr)
        # Flush and close sys.stderr - also closes the file descriptor (fd)
        sys.stderr.close()
        # Make original_stderr_fd point to the same file as to_fd
        os.dup2(to_fd, original_stderr_fd)
        # Create a new sys.stderr that points to the redirected fd
        sys.stderr = io.TextIOWrapper(os.fdopen(original_stderr_fd, 'wb'))

    # The original fd stderr points to. Usually 2 on POSIX systems.
    original_stderr_fd = sys.stderr.fileno()
    # Save a copy of the original stderr fd in saved_stderr_fd
    saved_stderr_fd = os.dup(original_stderr_fd)
    # Create a temporary file and redirect stderr to it
    tfile = tempfile.TemporaryFile(mode='w+b')
    try:
        _redirect_stderr(tfile.fileno())
        # Yield to caller, then redirect stderr back to the saved fd
        yield
        _redirect_stderr(saved_stderr_fd)
        # Copy contents of temporary file to the given stream
        tfile.flush()
        tfile.seek(0, io.SEEK_SET)
        stream.write(tfile.read())
    finally:
        tfile.close()
        os.close(saved_stderr_fd)



import sys
import timeit

import typing
from pymemtrace import debug_malloc_stats

def create_string(l: int) -> str:
    return ' ' * l


def test_under_512(count: int, list_of_strings: typing.List[str]):
    for i in range(count):
        list_of_strings.append(create_string(256))


def test_over_512(count: int, list_of_strings: typing.List[str]):
    for i in range(count):
        list_of_strings.append(create_string(1024))


def test_well_over_512(count: int, list_of_strings: typing.List[str]):
    for i in range(count):
        list_of_strings.append(create_string(1024**2))


def example_debug_malloc_stats_for_documentation(list_of_strings):
    """An example of using the trace_malloc.trace_malloc_report decorator for logging memory usage.
    Typical output:

    .. code-block:: text

        example_trace_malloc_for_documentation()
        DiffSysDebugMallocStats.diff():
        class   size   num pools   blocks in use  avail blocks
        -----   ----   ---------   -------------  ------------
            1     32          +1             +52           +74
            2     48          +0             +17           -17
            3     64          +0             +33           -33
            4     80          +1             +51            -1
            5     96          +2             +34           +50
            6    112          +0              +2            -2
            7    128          +0              +1            -1
           10    176          +0              +1            -1
           12    208          +0              +1            -1
           17    288          +0              +1            -1
           18    304          +0              +2            -2
           25    416          +0              +3            -3
           26    432          +0              +3            -3
           27    448          +0              +3            -3
           29    480          +0              +3            -3
           30    496          +0              +1            -1
           31    512          +0              +1            -1

        # bytes in allocated blocks        =              +19,904
        # bytes in available blocks        =               -3,808
        -4 unused pools * 4096 bytes       =              -16,384
        # bytes lost to pool headers       =                 +192
        # bytes lost to quantization       =                  +96

          -1 free 1-sized PyTupleObjects * 32 bytes each =                  -32
          +1 free 5-sized PyTupleObjects * 64 bytes each =                  +64
                   +2 free PyDictObjects * 48 bytes each =                  +96
                   -2 free PyListObjects * 40 bytes each =                  -80
                 +1 free PyMethodObjects * 48 bytes each =                  +48

    """
    print(f'example_trace_malloc_for_documentation()')
    with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
        for i in range(1, 9):
            list_of_strings.append(' ' * (i * 8))
    print(f'DiffSysDebugMallocStats.diff():')
    print(f'{malloc_diff.diff()}')


COUNT = 8

def example():
    for function in (test_under_512, test_over_512, test_well_over_512):
        print(f'Function: {function}'.center(75, '='))
        list_of_strings = []
        with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
            function(COUNT, list_of_strings)
        print(f'DiffSysDebugMallocStats.diff():')
        print(f'{malloc_diff.diff()}')
        print(f'DONE: Function: {function}'.center(75, '='))
        print()


def example_timeit():
    for function in (test_under_512, test_over_512, test_well_over_512):
        list_of_strings = []
        function(COUNT, list_of_strings)


def example_timeit_with_debug_malloc_stats():
    for function in (test_under_512, test_over_512, test_well_over_512):
        list_of_strings = []
        with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
            function(COUNT, list_of_strings)


def example_timeit_under_512():
    list_of_strings = []
    test_under_512(COUNT, list_of_strings)


def example_timeit_under_512_with_debug_malloc_stats():
    list_of_strings = []
    with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
        test_under_512(COUNT, list_of_strings)


def example_timeit_over_512():
    list_of_strings = []
    test_over_512(COUNT, list_of_strings)


def example_timeit_over_512_with_debug_malloc_stats():
    list_of_strings = []
    with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
        test_over_512(COUNT, list_of_strings)


def example_timeit_well_over_512():
    list_of_strings = []
    test_well_over_512(COUNT, list_of_strings)


def example_timeit_well_over_512_with_debug_malloc_stats():
    list_of_strings = []
    with debug_malloc_stats.DiffSysDebugMallocStats() as malloc_diff:
        test_well_over_512(COUNT, list_of_strings)


def main():
    # example_debug_malloc_stats_for_documentation([])
    print()
    # example()
    # print(timeit.repeat('p.memory_info().rss', setup='import psutil; p = psutil.Process()', number=1_000_000, repeat=5))

    NUMBER = 10_000
    REPEAT = 5
    CONVERT = 1_000_000
    print(f'number={NUMBER:,d} repeat={REPEAT:,d} convert={CONVERT:,d}')
    for function in (
            'example_timeit_under_512',
            'example_timeit_over_512',
            'example_timeit_well_over_512',
    ):
        # t = timeit.timeit(f"{function}()", setup=f"from __main__ import {function}", number=NUMBER) / NUMBER
        # print(f'{function:60}: {t:9.9f}')
        times = timeit.repeat(f"{function}()", setup=f"from __main__ import {function}", number=NUMBER, repeat=REPEAT)
        times = [CONVERT * t / NUMBER for t in times]
        result = [f'{v:9.3f}' for v in times]
        times_mean = sum(times) / REPEAT
        print(
            f'{function:60}:'
            f' {", ".join(result)}'
            f' mean={times_mean:9.3f}'
            f' min={min(times):9.3f}'
            f' max={max(times):9.3f}'
            f' span={max(times) - min(times):9.3f}'
        )
        # With _with_debug_malloc_stats
        times_with_debug_malloc_stats = timeit.repeat(f"{function}_with_debug_malloc_stats()", setup=f"from __main__ import {function}_with_debug_malloc_stats", number=NUMBER, repeat=REPEAT)
        times_with_debug_malloc_stats = [CONVERT * t / NUMBER for t in times_with_debug_malloc_stats]
        result = [f'{v:9.3f}' for v in times_with_debug_malloc_stats]
        times_mean_with_debug_malloc_stats = sum(times_with_debug_malloc_stats) / REPEAT
        print(
            f'{function + "_with_debug_malloc_stats":60}:'
            f' {", ".join(result)}'
            f' mean={times_mean_with_debug_malloc_stats:9.3f}'
            f' min={min(times_with_debug_malloc_stats):9.3f}'
            f' max={max(times_with_debug_malloc_stats):9.3f}'
            f' span={max(times_with_debug_malloc_stats) - min(times_with_debug_malloc_stats):9.3f}'
            f' x{times_mean_with_debug_malloc_stats / times_mean:>8.3f}'
        )
    return 0


if __name__ == '__main__':
    sys.exit(main())




"""
This is a wrapper around sys._debugmallocstats which writes to C stderr.
We capture this and can diff two calls to sys._debugmallocatats.

sys._debugmallocatats Implementation (Python 3.9)
=================================================

``sys__debugmallocstats_impl(PyObject *module)``

Calls:

``_PyObject_DebugMallocStats(stderr))``

Then:

``_PyObject_DebugTypeStats(stderr);``

Memory Usage by the Python Object Allocator
---------------------------------------------

``_PyObject_DebugMallocStats is defined in Objects/obmalloc.c:``

Which dumps out the arenas/pools/blocks.

Memory Usage by Type
------------------------------

``_PyObject_DebugTypeStats`` is defined in ``Objects/object.c``

This calls::

    _PyDict_DebugMallocStats(out); // in Objects/dictobject.c, just calls _PyDebugAllocatorStats in Objects/obmalloc.c
    _PyFloat_DebugMallocStats(out); // in Objects/dictobject.c, just calls _PyDebugAllocatorStats in Objects/obmalloc.c
    _PyFrame_DebugMallocStats(out); // etc.
    _PyList_DebugMallocStats(out);
    _PyTuple_DebugMallocStats(out); // in Objects/tupleobject.c, calls _PyDebugAllocatorStats in Objects/obmalloc.c for (i = 1; i < PyTuple_MAXSAVESIZE; i++)

Note that only dict, float, frame, list, tuple are reported.
"""
import io
import re
import sys
import typing

from pymemtrace import redirect_stdout


def get_debugmallocstats() -> bytes:
    """Invokes sys._debugmallocstats and captures the output as bytes."""
    stream = io.BytesIO()
    with redirect_stdout.stderr_redirector(stream):
        sys._debugmallocstats()
    return stream.getvalue()
    # return stream.getvalue().decode('ascii', 'replace')


#: In ``Object/obmalloc.c``:
#:
#: .. code-block:: text
#:
#:      #define POOL_SIZE               SYSTEM_PAGE_SIZE        /* must be 2^N */
#:
POOL_SIZE = 4096

#: This value is initially approximate.
#: In ``Object/obmalloc.c``:
#:
#: .. code-block:: c
#:
#:      #define POOL_OVERHEAD   _Py_SIZE_ROUND_UP(sizeof(struct pool_header), ALIGNMENT)
#:
#: We can calculate this from the sum of num_pools divided into ``'# bytes lost to pool headers'``.
#: This is done whenever a :py:class:`SysDebugMallocStats` is created.
POOL_OVERHEAD = 48


class DebugMallocStat(typing.NamedTuple):
    """Represents a single line in the malloc stats section. For example:

    .. code-block:: text

        class   size   num pools   blocks in use  avail blocks
        -----   ----   ---------   -------------  ------------
            0     16           2             297           209

    Nomenclature is from ``_PyObject_DebugMallocStats(stderr))`` in ``Objects/obmalloc.c``.
    Typical implementation:

    .. code-block:: c

        for (i = 0; i < numclasses; ++i) {
            size_t p = numpools[i];
            size_t b = numblocks[i];
            size_t f = numfreeblocks[i];
            uint size = INDEX2SIZE(i);
            if (p == 0) {
                assert(b == 0 && f == 0);
                continue;
            }
            fprintf(out, "%5u %6u "
                            "%11" PY_FORMAT_SIZE_T "u "
                            "%15" PY_FORMAT_SIZE_T "u "
                            "%13" PY_FORMAT_SIZE_T "u\\n",
                    i, size, p, b, f);
            allocated_bytes += b * size;
            available_bytes += f * size;
            pool_header_bytes += p * POOL_OVERHEAD;
            quantization += p * ((POOL_SIZE - POOL_OVERHEAD) % size);
        }
        fputc('\\n', out);

    """
    block_class: int
    size: int
    num_pools: int
    blocks_in_use: int
    avail_blocks: int

    # @property
    # def bytes_in_use(self)-> int:
    #     return self.size * self.num_pools * self.blocks_in_use

    @property
    def allocated_bytes(self) -> int:
        return self.blocks_in_use * self.size

    @property
    def available_bytes(self) -> int:
        return self.avail_blocks * self.size

    @property
    def pool_header_bytes(self) -> int:
        return self.num_pools * POOL_OVERHEAD

    @property
    def quantization(self) -> int:
        return self.num_pools * ((POOL_SIZE - POOL_OVERHEAD) % self.size)

    def __repr__(self):
        """Representation of self of the form:

        .. code-block:: text

                0     16           4             777           235

        """
        return f'{self.block_class:5d}  {self.size:5d}  {self.num_pools:10d}  {self.blocks_in_use:14d}  {self.avail_blocks:12d}'


def diff_debug_malloc_stat(a: DebugMallocStat, b: DebugMallocStat) -> str:
    """Takes two DebugMallocStat objects and returns a string with the difference.
    The string is of similar format to the input from ``sys._debugmallocstats``."""
    if a.block_class != b.block_class:
        raise ValueError(f'a.block_class != b.block_class: {a.block_class} != {b.block_class}')
    if a.size != b.size:
        raise ValueError(f'a.size != b.size: {a.size} != {b.size}')
    return (
        f'{a.block_class:5d}'
        f'  {a.size:5d}'
        f'  {b.num_pools - a.num_pools:+10d}'
        f'  {b.blocks_in_use - a.blocks_in_use:+14d}'
        f'  {b.avail_blocks - a.avail_blocks:+12d}')


#: Matches::
#:
#:      b'    0     16           2             294           212'
#:
#: Decomposed to extract the five integers.
RE_DEBUG_MALLOC_STATS_LINE = re.compile(rb'^\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$')


def last_value_as_int(line: bytes) -> int:
    """Returns that last value of the line such as:

    .. code-block:: text

        b'# arenas allocated total           =                2,033'

    """
    return int(line.split(b'=')[1].replace(b',', b''))


#: Matches::
#:
#:      b'18 arenas * 262144 bytes/arena     =            4,718,592'
#:
#: Decomposed to extract the two integers, the total is computed.
RE_DEBUG_MALLOC_ARENAS_SUMMARY_LINE = re.compile(rb'^(\d+) arenas \* (\d+) bytes/arena\s+=\s+(.+)$')


class DebugMallocArenas:
    """Decomposes this::

        # arenas allocated total           =                2,033
        # arenas reclaimed                 =                2,015
        # arenas highwater mark            =                   18
        # arenas allocated current         =                   18
        18 arenas * 262144 bytes/arena     =            4,718,592

    Into values for: ntimes_arena_allocated, narenas, narenas_highwater, ARENA_SIZE.

    Infers: arenas_reclaimed, arenas_total.

    From Object/obmalloc.c:

    .. code-block:: c

        (void)printone(out, "# arenas allocated total", ntimes_arena_allocated);
        (void)printone(out, "# arenas reclaimed", ntimes_arena_allocated - narenas);
        (void)printone(out, "# arenas highwater mark", narenas_highwater);
        (void)printone(out, "# arenas allocated current", narenas);

        /* Total number of times malloc() called to allocate an arena. */
        static size_t ntimes_arena_allocated = 0;
        // b'# arenas allocated total           =                2,033'

        /* High water mark (max value ever seen) for narenas_currently_allocated. */
        static size_t narenas_highwater = 0;
        // b'# arenas highwater mark            =                   18'

        /* # of arenas actually allocated. */
        size_t narenas = 0;
        // b'# arenas allocated current         =                   18'

        PyOS_snprintf(buf, sizeof(buf), "%" PY_FORMAT_SIZE_T "u arenas * %d bytes/arena", narenas, ARENA_SIZE);
        (void)printone(out, buf, narenas * ARENA_SIZE);
        // b'18 arenas * 262144 bytes/arena     =            4,718,592'
        // Simple calculation: 18 * 262144 = 4718592

    """
    def __init__(self, debug_malloc: bytes):
        """Constructor, decomposes this:

        .. code-block:: text

            # arenas allocated total        -> self.ntimes_arena_allocated
            # arenas reclaimed              -> self.arenas_reclaimed
            # arenas highwater mark         -> self.narenas_highwater
            # arenas allocated current      -> self.narenas
            18 arenas * 262144 bytes/arena  -> self.narenas * self.arena_size  = self.arenas_total

        """
        for line in debug_malloc.splitlines(keepends=False):
            if line.startswith(b'# arenas allocated total'):
                # (void)printone(out, "# arenas allocated total", ntimes_arena_allocated);
                self.ntimes_arena_allocated = last_value_as_int(line)
            elif line.startswith(b'# arenas reclaimed'):
                # (void)printone(out, "# arenas reclaimed", ntimes_arena_allocated - narenas);
                pass
            elif line.startswith(b'# arenas highwater mark'):
                # (void)printone(out, "# arenas highwater mark", narenas_highwater);
                self.narenas_highwater = last_value_as_int(line)
            elif line.startswith(b'# arenas allocated current'):
                # (void)printone(out, "# arenas allocated current", narenas);
                self.narenas = last_value_as_int(line)
            else:
                m = RE_DEBUG_MALLOC_ARENAS_SUMMARY_LINE.match(line)
                if m:
                    # PyOS_snprintf(buf, sizeof(buf), "%" PY_FORMAT_SIZE_T "u arenas * %d bytes/arena", narenas, ARENA_SIZE);
                    # (void)printone(out, buf, narenas * ARENA_SIZE);
                    #
                    # b'18 arenas * 262144 bytes/arena     =            4,718,592'
                    # Simple calculation: 18 * 262144 = 4718592
                    self.arena_size = int(m.group(2))
        expected_attrs = ('ntimes_arena_allocated', 'narenas_highwater', 'narenas', 'arena_size')
        if not all(hasattr(self, name) for name in expected_attrs):
            raise ValueError(f'Can not find required attributes {expected_attrs}')

    @property
    def arenas_reclaimed(self) -> int:
        return self.ntimes_arena_allocated - self.narenas

    @property
    def arenas_total(self) -> int:
        return self.narenas * self.arena_size

    def __repr__(self):
        """Returns a string similar to sys._debugmallocstats."""
        ret= [
                f'# arenas allocated total           = {self.ntimes_arena_allocated:>20,d}',
                f'# arenas reclaimed                 = {self.arenas_reclaimed:>20,d}',
                f'# arenas highwater mark            = {self.narenas_highwater:>20,d}',
                f'# arenas allocated current         = {self.narenas:>20,d}',
            ]
        lhs = f'{self.narenas} arenas * {self.arena_size} bytes/arena'
        ret.append(f'{lhs:<34s} = {self.arenas_total:>20,d}')
        return '\n'.join(ret)


#: Matches::
#:
#:      b'63 unused pools * 4096 bytes       =              258,048'
#:
#: Decomposed to extract the two integers, the total is computed.
RE_DEBUG_MALLOC_POOLS_SUMMARY_LINE = re.compile(rb'^(\d+) unused pools \* (\d+) bytes\s+=\s+(.+)$')


class DebugMallocPoolsBlocks:
    """Decomposes this::

        # bytes in allocated blocks        =            4,280,848
        # bytes in available blocks        =               70,368
        63 unused pools * 4096 bytes       =              258,048
        # bytes lost to pool headers       =               52,272
        # bytes lost to quantization       =               57,056
        # bytes lost to arena alignment    =                    0
        Total                              =            4,718,592

    From Object/obmalloc.c::

        total = printone(out, "# bytes in allocated blocks", allocated_bytes);
        total += printone(out, "# bytes in available blocks", available_bytes);

        PyOS_snprintf(buf, sizeof(buf), "%u unused pools * %d bytes", numfreepools, POOL_SIZE);
        total += printone(out, buf, (size_t)numfreepools * POOL_SIZE);

        total += printone(out, "# bytes lost to pool headers", pool_header_bytes);
        total += printone(out, "# bytes lost to quantization", quantization);
        total += printone(out, "# bytes lost to arena alignment", arena_alignment);
        (void)printone(out, "Total", total);

    Extracts: allocated_bytes, available_bytes, numfreepools, POOL_SIZE, pool_header_bytes, quantization, arena_alignment.

    Infers: unused_pool_total, TOTAL.
    """
    def __init__(self, debug_malloc: bytes):
        for line in debug_malloc.splitlines(keepends=False):
            if line.startswith(b'# bytes in allocated blocks'):
                # total = printone(out, "# bytes in allocated blocks", allocated_bytes);
                self.allocated_bytes = last_value_as_int(line)
            elif line.startswith(b'# bytes in available blocks'):
                # total += printone(out, "# bytes in available blocks", available_bytes);
                self.available_bytes = last_value_as_int(line)
            elif line.startswith(b'# bytes lost to pool headers'):
                # total += printone(out, "# bytes lost to pool headers", pool_header_bytes);
                self.pool_header_bytes = last_value_as_int(line)
            elif line.startswith(b'# bytes lost to quantization'):
                # total += printone(out, "# bytes lost to quantization", quantization);
                self.quantization = last_value_as_int(line)
            elif line.startswith(b'# bytes lost to arena alignment'):
                # total += printone(out, "# bytes lost to arena alignment", arena_alignment);
                self.arena_alignment = last_value_as_int(line)
            else:
                m = RE_DEBUG_MALLOC_POOLS_SUMMARY_LINE.match(line)
                if m:
                    # PyOS_snprintf(buf, sizeof(buf), "%u unused pools * %d bytes", numfreepools, POOL_SIZE);
                    # total += printone(out, buf, (size_t)numfreepools * POOL_SIZE);
                    #
                    # 63 unused pools * 4096 bytes       =              258,048
                    # Simple calculation: 63 * 4096 = 258048
                    self.numfreepools = int(m.group(1))
                    self.pool_size = int(m.group(2))
        expected_attrs = (
        'allocated_bytes', 'available_bytes', 'pool_header_bytes', 'quantization', 'arena_alignment', 'numfreepools',
        'pool_size'
        )
        if not all(hasattr(self, name) for name in expected_attrs):
            raise ValueError(f'Can not find required attributes {expected_attrs}')

    @property
    def unused_pool_total(self) -> int:
        return self.numfreepools * self.pool_size

    @property
    def total(self) -> int:
        return self.allocated_bytes + self.available_bytes + self.unused_pool_total + self.pool_header_bytes \
               + self.quantization + self.arena_alignment

    def pool_overhead(self, num_pools: int) -> int:
        """Returns the POOL_OVERHEAD as self.pool_header_bytes // the number of pools.

        self.pool_header_bytes comes from::

            # bytes lost to pool headers       =               51,264

        Number of pools comes from the sum of ``num pools`` from DebugMallocStat.num_pools
        """
        return self.pool_header_bytes // num_pools

    def __repr__(self):
        """Returns a string similar to sys._debugmallocstats."""
        ret = [
            f'# bytes in allocated blocks        = {self.allocated_bytes:>20,d}',
            f'# bytes in available blocks        = {self.available_bytes:>20,d}',
        ]
        lhs = f'{self.numfreepools} unused pools * {self.pool_size} bytes'
        ret.append(f'{lhs:<34s} = {self.unused_pool_total:>20,d}')
        ret.extend([
            f'# bytes lost to pool headers       = {self.pool_header_bytes:>20,d}',
            f'# bytes lost to quantization       = {self.quantization:>20,d}',
            f'# bytes lost to arena alignment    = {self.arena_alignment:>20,d}',
            f'Total                              = {self.total:>20,d}',
        ])
        return '\n'.join(ret)


class DebugTypeStat(typing.NamedTuple):
    """Represents a single line from ``sys._debugmallocstats``.

    Decomposed from a line such as:

    .. code-block:: text

        4 free PyCFunctionObjects * 56 bytes each =                  224

    See ``_PyObject_DebugTypeStats(stderr);`` in ``Objects/obmalloc.c``
    """
    free_count: int
    object_type: str
    bytes_each: int
    bytes_total: int

    def __repr__(self):
        """Returns a string of the form of these lines:

        .. code-block:: text

                   4 free PyCFunctionObjects * 56 bytes each =                  224
                        9 free PyDictObjects * 48 bytes each =                  432
                       5 free PyFloatObjects * 24 bytes each =                  120
                      0 free PyFrameObjects * 368 bytes each =                    0
                       80 free PyListObjects * 40 bytes each =                3,200
                      8 free PyMethodObjects * 48 bytes each =                  384
               7 free 1-sized PyTupleObjects * 32 bytes each =                  224
              52 free 2-sized PyTupleObjects * 40 bytes each =                2,080
               1 free 3-sized PyTupleObjects * 48 bytes each =                   48
             0 free 10-sized PyTupleObjects * 104 bytes each =                    0

        """
        lhs = f'{self.free_count} free {self.object_type} * {self.bytes_each} bytes each'
        return f'{lhs:>48s} = {self.bytes_total:>20,d}'


#: Matches::
#:
#:      b'   34 free 2-sized PyTupleObjects * 40 bytes each =                1,360'
#:
#: NOTE: commas in last value caused by printone() in Object/obmalloc.c
#: printone is used in many places where there is message = value such as memory pool totals and type memory information.
RE_DEBUG_MALLOC_TYPE_LINE = re.compile(rb'^\s*(\d+) free (.+?) \* (\d+) bytes each =\s+(.+)$')


def diff_debug_type_stat(a: DebugTypeStat, b: DebugTypeStat) -> str:
    """Takes two DebugMallocStat objects and returns a string with the difference.
    The string is of similar format to the input from ``sys._debugmallocstats``."""
    if a.object_type != b.object_type:
        raise ValueError(f'a.object_type != b.object_type: {a.object_type} != {b.object_type}')
    if a.bytes_each != b.bytes_each:
        raise ValueError(f'a.bytes_each != b.bytes_each: {a.bytes_each} != {b.bytes_each}')
    lhs = f'{b.free_count - a.free_count:+d} free {a.object_type} * {a.bytes_each} bytes each'
    return f'{lhs:>48s} = {b.bytes_total - a.bytes_total:>+20,d}'


#: Matches::
#:
#:      b'Small block threshold = 512, in 32 size classes.'
#:
#: Decomposed to extract the two integers.
RE_DEBUG_MALLOC_HEADER_LINE = re.compile(rb'^Small block threshold = (\d+), in (\d+) size classes\.$')


class SysDebugMallocStats:
    """This decomposes the output of ``sys._debugmallocstats`` into these areas:

    - A list of malloc stats showing the pools and blocks.
    - Descriptions of arenas.
    - Descriptions of pools and blocks.
    - A list of malloc usage by (some) types.

    This class takes a snapshot of the debug malloc stats from ``sys._debugmallocstats``.
    Importantly it can identify the difference between two snapshots.
    """
    def __init__(self, debug_malloc: bytes = b''):
        """Constructor, this optionally takes a bytes object for testing.
        If nothing supplied this gets the bytes object from sys._debugmallocstats.
        """
        self.malloc_stats: typing.List[DebugMallocStat] = []
        self.type_stats: typing.List[DebugTypeStat] = []
        # Used for lookup by type
        self.type_map: typing.Dict[bytes, int] = {}
        if not debug_malloc:
            debug_malloc: bytes = get_debugmallocstats()
        for line in debug_malloc.splitlines(keepends=False):
            m = RE_DEBUG_MALLOC_STATS_LINE.match(line)
            if m:
                self.malloc_stats.append(DebugMallocStat(*[int(v) for v in m.groups()]))
            else:
                m = RE_DEBUG_MALLOC_TYPE_LINE.match(line)
                if m:
                    self.type_map[m.group(2)] = len(self.type_stats)
                    self.type_stats.append(DebugTypeStat(
                            int(m.group(1)), m.group(2).decode('ascii'), int(m.group(3)),
                            int(m.group(4).replace(b',', b''))
                        ))
                else:
                    m = RE_DEBUG_MALLOC_HEADER_LINE.match(line)
                    if m:
                        self.small_block_threshold = int(m.group(1))
                        self.size_classes = int(m.group(2))
        self.arenas = DebugMallocArenas(debug_malloc)
        self.pools_blocks = DebugMallocPoolsBlocks(debug_malloc)
        expected_attrs = ('small_block_threshold', 'size_classes')
        if not all(hasattr(self, name) for name in expected_attrs):
            raise ValueError(f'Can not find required attributes {expected_attrs}')
        # Set the global POOL_OVERHEAD by combining malloc_stats and pools_blocks
        num_pools = sum(v.num_pools for v in self.malloc_stats)
        pool_overhead = self.pools_blocks.pool_overhead(num_pools)
        global POOL_OVERHEAD
        POOL_OVERHEAD = pool_overhead

    def object_types(self) -> typing.KeysView[bytes]:
        """Return all the known object types."""
        return self.type_map.keys()

    def has_object_type(self, object_type: bytes):
        """Return True if the object type is present."""
        return object_type in self.type_map

    def type_stat(self, object_type: bytes) -> DebugTypeStat:
        """Return the DebugTypeStat for the named object type.
        May raise an KeyError if the object_type doe not exist."""
        return self.type_stats[self.type_map[object_type]]

    def __repr__(self):
        """Representation of self similar to the output of sys._debugmallocstats"""
        results = [
            f'Small block threshold = {self.small_block_threshold}, in {self.size_classes} size classes.',
            '',
            'class   size   num pools   blocks in use  avail blocks',
            '-----   ----   ---------   -------------  ------------',
        ]
        for stat in self.malloc_stats:
            results.append(repr(stat))
        results.append('')
        results.append(repr(self.arenas))
        results.append('')
        results.append(repr(self.pools_blocks))
        results.append('')
        for stat in self.type_stats:
            results.append(repr(stat))
        # Terminate with a '\n'
        results.append('')
        return '\n'.join(results)


# def debugmallocstats_to_malloc_blocks() -> typing.List[DebugMallocStats]:
#     debug_malloc: bytes = get_debugmallocstats()
#     ret = []
#     for line in debug_malloc.splitlines(keepends=False):
#         m = RE_DEBUG_MALLOC_STATS_LINE.match(line)
#         if m:
#             ret.append(DebugMallocStats(*[int(v) for v in m.groups()]))
#     return ret


def diff_debugmallocstats(a_stats: SysDebugMallocStats, b_stats: SysDebugMallocStats):
    """
    This takes two SysDebugMallocStats objects and identifies what is different between them.
    The diff is a list of lines of identical form to :py:meth:`sys._debugmallocstats` with '+' or '-' where appropriate.
    Lines that are the same are omitted.
    """
    ret: typing.List[str] = []
    has_header = False

    # Firstly the DebugMallocStat list
    if len(a_stats.malloc_stats) != len(b_stats.malloc_stats):
        raise ValueError(
            f'Malloc stats length  mismatch {len(a_stats.malloc_stats)} != {len(b_stats.malloc_stats)}'
        )
    for a, b in zip(a_stats.malloc_stats, b_stats.malloc_stats):
        if a != b:
            if not has_header:
                ret.append('class   size   num pools   blocks in use  avail blocks')
                ret.append('-----   ----   ---------   -------------  ------------')
                has_header = True
            ret.append(diff_debug_malloc_stat(a, b))
    has_header = False
    # Central two blocks
    # Arenas
    block = []
    if b_stats.arenas.ntimes_arena_allocated != a_stats.arenas.ntimes_arena_allocated:
        block.append(
            f'# arenas allocated total           ='
            f' {b_stats.arenas.ntimes_arena_allocated - a_stats.arenas.ntimes_arena_allocated:>+20,d}'
        )
    if b_stats.arenas.arenas_reclaimed != a_stats.arenas.arenas_reclaimed:
        block.append(
            f'# arenas reclaimed                 ='
            f' {b_stats.arenas.arenas_reclaimed - a_stats.arenas.arenas_reclaimed:>+20,d}'
        )
    if b_stats.arenas.narenas_highwater != a_stats.arenas.narenas_highwater:
        block.append(
            f'# arenas highwater mark            ='
            f' {b_stats.arenas.narenas_highwater - a_stats.arenas.narenas_highwater:>+20,d}'
        )
    assert a_stats.arenas.arena_size == b_stats.arenas.arena_size
    if b_stats.arenas.narenas != a_stats.arenas.narenas:
        block.append(
            f'# arenas allocated current         ='
            f' {b_stats.arenas.narenas - a_stats.arenas.narenas:>+20,d}'
        )
        lhs = f'{b_stats.arenas.narenas - a_stats.arenas.narenas:+d} arenas * {a_stats.arenas.arena_size} bytes/arena'
        block.append(f'{lhs:<34s} = {b_stats.arenas.arenas_total - a_stats.arenas.arenas_total:>+20,d}')
    if block:
        ret.append('')
        ret.extend(block)
    # Pools and blocks
    block = []
    assert a_stats.pools_blocks.pool_size == b_stats.pools_blocks.pool_size
    if b_stats.pools_blocks.allocated_bytes != a_stats.pools_blocks.allocated_bytes:
        block.append(
            f'# bytes in allocated blocks        ='
            f' {b_stats.pools_blocks.allocated_bytes - a_stats.pools_blocks.allocated_bytes:>+20,d}'
        )
    if b_stats.pools_blocks.available_bytes != a_stats.pools_blocks.available_bytes:
        block.append(
            f'# bytes in available blocks        ='
            f' {b_stats.pools_blocks.available_bytes - a_stats.pools_blocks.available_bytes:>+20,d}'
        )
    if b_stats.pools_blocks.numfreepools != a_stats.pools_blocks.numfreepools:
        lhs = (
            f'{b_stats.pools_blocks.numfreepools - a_stats.pools_blocks.numfreepools}'
            f' unused pools * {a_stats.pools_blocks.pool_size} bytes'
        )
        block.append(
            f'{lhs:<34s} = {b_stats.pools_blocks.unused_pool_total - a_stats.pools_blocks.unused_pool_total:>+20,d}'
        )
    if b_stats.pools_blocks.pool_header_bytes != a_stats.pools_blocks.pool_header_bytes:
        block.append(
            f'# bytes lost to pool headers       ='
            f' {b_stats.pools_blocks.pool_header_bytes - a_stats.pools_blocks.pool_header_bytes:>+20,d}'
        )
    if b_stats.pools_blocks.quantization != a_stats.pools_blocks.quantization:
        block.append(
            f'# bytes lost to quantization       ='
            f' {b_stats.pools_blocks.quantization - a_stats.pools_blocks.quantization:>+20,d}'
        )
    if b_stats.pools_blocks.arena_alignment != a_stats.pools_blocks.arena_alignment:
        block.append(
            f'# bytes lost to arena alignment    ='
            f' {b_stats.pools_blocks.arena_alignment - a_stats.pools_blocks.arena_alignment:>+20,d}'
        )
    if b_stats.pools_blocks.total != a_stats.pools_blocks.total:
        block.append(
            f'Total                              = {b_stats.pools_blocks.total - a_stats.pools_blocks.total:>+20,d}'
        )
    if block:
        ret.append('')
        ret.extend(block)
    has_header = False
    # The DebugTypeStat objects, the lists might not be the same.
    union_object_types = set(a_stats.object_types()) | set(b_stats.object_types())
    for object_type in sorted(union_object_types):
        diff_str = ''
        if a_stats.has_object_type(object_type):
            if b_stats.has_object_type(object_type):
                if a_stats.type_stat(object_type) != b_stats.type_stat(object_type):
                    diff_str = diff_debug_type_stat(a_stats.type_stat(object_type), b_stats.type_stat(object_type))
            else:
                # a only so dropped
                diff_str = f'-{a_stats.type_stat(object_type)!r}'
        else:
            # b only so added
            diff_str = f'+{a_stats.type_stat(object_type)!r}'
        if diff_str:
            if not has_header:
                ret.append('')
                has_header = True
            ret.append(diff_str)
    return ret


class DiffSysDebugMallocStats:
    """Context manager that compares two snapshots of ``sys._getdebugmallocstats()`` and can provide a diff between
    them."""
    def __init__(self):
        self.before: typing.Optional[SysDebugMallocStats] = None
        self.after: typing.Optional[SysDebugMallocStats] = None
        self._diff: typing.Optional[str] = None

    def __enter__(self):
        """Enters the context manager taking a snapshot of ``sys._getdebugmallocstats()``."""
        self.before = SysDebugMallocStats()
        self.after = None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exits the context manager taking a snapshot of ``sys._getdebugmallocstats()``."""
        self.after = SysDebugMallocStats()
        return False

    def diff(self) -> str:
        """Returns the difference between two snapshots."""
        if self.before is None:
            raise RuntimeError('Context manager not entered.')
        if self.after is None:
            raise RuntimeError('Context manager not exited.')
        return '\n'.join(diff_debugmallocstats(self.before, self.after))


def main():
    print('sys._debugmallocstats()')
    print(get_debugmallocstats().decode('ascii'))

    print()
    # dms_a = SysDebugMallocStats()
    # # print(repr(dms_a))
    # l = []
    # l.append({})
    # l.append(set())
    # l.append((1, 2, 3))
    # for i in range(80):
    #     l.append(tuple(list(range(4))))
    # dms_b = SysDebugMallocStats()
    #
    # # pprint.pprint(diff_debugmallocstats(dms_a, dms_b))
    # print('\n'.join(diff_debugmallocstats(dms_a, dms_b)))

    print(f'POOL_OVERHEAD {POOL_OVERHEAD}')

    with DiffSysDebugMallocStats() as diff_dms:
        l = []
        l.append({})
        l.append(set())
        l.append((1, 2, 3))
        for i in range(80):
            l.append(tuple(list(range(4))))
    print(f'POOL_OVERHEAD {POOL_OVERHEAD}')
    print(diff_dms.diff())

    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Test the behaviour of sys.setrace() to understand the sequence and type of
events produced by various function and generator calls.

Created on 6 Dec 2017

@author: paulross
"""
import inspect
import collections
import pprint
import sys

import pytest

TRACE_TO_STDOUT = False

# RecordTraceEvents accumulates a list of these
TraceRecord = collections.namedtuple(
    'TraceRecord', 'filename, function, lineno, lineno_code, event, arg'
)
TraceRecord.__doc__ += ': Data from an event seen by sys.setprofile() or sys.settrace()'
TraceRecord.filename.__doc__ = 'Absolute file path as a string.' 
TraceRecord.function.__doc__ = 'Function (unqualified) name as a string.' 
TraceRecord.lineno.__doc__ = 'Line number of the event as an int.'
TraceRecord.lineno_code.__doc__ = 'Line number of the start of the function' \
    ' or generator where the event is as an int. This is constant per function' \
    ' regardless of where the return or yield statements is.'
TraceRecord.event.__doc__ = "Event type: ('call', 'line', 'return', 'exception', 'c_call', 'c_return', or 'c_exception')."
TraceRecord.arg.__doc__ = "Argument: 'call': always None, 'line': always None," \
    " 'return' return value. 'exception': (exception, value, traceback)" \
    "'c_call', 'c_return', 'c_exception': the C function object."


class RecordTraceEvents:
    """
    Class to record events generated by sys.settrace().
    """
    def __init__(self, events_to_trace=None, ignore_self_exit=True):
        self.events_to_trace = events_to_trace
        self.records = []
        self._trace_fn_stack = []
        self._profile_fn_stack = []
        if ignore_self_exit:
            self._call__exit__ = (__file__, self.__exit__lineno, '__exit__')
        else:
            self._call__exit__ = None
        
    def __str__(self):
#         return '\n'.join(
#             ['{:30} {:4d} {:4d} {:10s} {:s}'.format(
#                 r.function, r.lineno, r.lineno_code, r.event, '{!r:s}'.format(r.arg) if r.event != 'call' else '')
#                     for r in self.records
#              ]
#         )
        result = []
        for r in self.records:
            if r.event == 'call':
                # Always None
                arg = ''
            else:
                arg = '{!r:s}'.format(r.arg)
            result.append(
                '{:30} {:4d} {:4d} {:10s} {:s}'.format(
                    r.function, r.lineno, r.lineno_code, r.event, arg
                )
            )
        return '\n'.join(result)
            
    def __call__(self, frame, event, arg):
        if self.events_to_trace is None or event in self.events_to_trace:
            # Traceback(filename, lineno, function, code_context, index)
            frame_info = inspect.getframeinfo(frame)
            tr = TraceRecord(
                frame_info.filename,
                frame_info.function,
                frame_info.lineno,
                frame.f_code.co_firstlineno,
                event,
                arg,
            )
            if self._call__exit__ is None \
            or (tr.filename, tr.lineno, tr.function) != self._call__exit__:
                self.records.append(tr)
        return self

    def __enter__(self):
        """
        Context manager sets the profiling function. This saves the existing
        tracing function.
        """
        self._trace_fn_stack.append(sys.gettrace())
        sys.settrace(self)
#         self._profile_fn_stack.append(sys.getprofile())
#         sys.setprofile(self)
        return self
    
    __exit__lineno = inspect.currentframe().f_lineno + 1 
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the context manager. This restores the tracing function to that
        prior to ``__enter__``.
        """
        sys.settrace(self._trace_fn_stack.pop())
#         sys.setprofile(self._profile_fn_stack.pop())

def test_simple_function():
    def simple_function_double(n):
        return n * 2

    with RecordTraceEvents(('call', 'return')) as rte:
        assert simple_function_double(14) == 28
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 2

def test_simple_function_multiple_returns():
    def function_multiple_returns(n):
        if n > 0:
            return 1
        elif n < 0:
            return -1
        return 0

    with RecordTraceEvents(('call', 'return')) as rte:
        assert function_multiple_returns(14) == 1
        assert function_multiple_returns(-14) == -1
        assert function_multiple_returns(0) == 0
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 6

def test_function_stack_two():
    def simple_function_child(n):
        return n * 2

    def simple_function_parent(n):
        return simple_function_child(2 * n)

    with RecordTraceEvents(('call', 'return')) as rte:
        assert simple_function_parent(1) == 4
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 4

def test_function_stack_two_decl_swap():
    def simple_function_parent(n):
        return simple_function_child(2 * n)

    def simple_function_child(n):
        return n * 2

    with RecordTraceEvents(('call', 'return')) as rte:
        assert simple_function_parent(1) == 4
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 4

def test_function_stack_three():
    def simple_function_child(n):
        return n * 2

    def simple_function_parent(n):
        return simple_function_child(2 * n)

    def simple_function_grandparent(n):
        return simple_function_parent(2 * n)

    with RecordTraceEvents(('call', 'return')) as rte:
        assert simple_function_grandparent(1) == 8
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 6

def test_function_recursive():
    def simple_recursive_function(n):
        assert n >= 0
        if n == 0:
            return 1
        return 2 * simple_recursive_function(n - 1)

    with RecordTraceEvents(('call', 'return')) as rte:
        assert simple_recursive_function(3) == 8
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 8
    
def test_same_named_functions():
    """Two classes that have the same named function called independently."""
    class NamedFunctionOne:
        def same_function(self, n):
            return n * 2

    class NamedFunctionTwo:
        def same_function(self, n):
            return n * 4

    with RecordTraceEvents(('call', 'return')) as rte:
        one = NamedFunctionOne()
        assert one.same_function(2) == 4
        two = NamedFunctionTwo()
        assert two.same_function(2) == 8
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 4

def test_same_named_functions_interleave():
    """Two classes that have the same named function called one after the other."""
    class NamedFunctionOne:
        def same_function(self, n):
            return n * 2

    class NamedFunctionTwo:
        def same_function(self, n):
            one = NamedFunctionOne()
            return one.same_function(n) * 4

    with RecordTraceEvents(('call', 'return')) as rte:
        two = NamedFunctionTwo()
        assert two.same_function(2) == 16
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 4

def test_generator():
    def simple_generator(n):
        for i in range(n):
            yield i

    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
#         results = [v for v in simple_generator(3)]
        results = []
        for v in simple_generator(3):
            results.append(v)
        assert results == [0, 1, 2]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 8

def test_generator_unfinished():
    def simple_generator(n):
        for i in range(n):
            yield i

    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
#         results = [v for v in simple_generator(3)]
        results = []
        gen = simple_generator(3)
        for i in range(2):
            results.append(next(gen))
        assert results == [0, 1]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert len(rte.records) == 4

def test_generator_parent_child():
    def child_generator(n):
        for i in range(n):
            yield i

    def parent_generator(n):
        for value in child_generator(n):
            yield 2 * value

    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
#         results = [v for v in simple_generator(3)]
        results = []
        for v in parent_generator(3):
            results.append(v)
        assert results == [0, 2, 4]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert sys.version_info.major == 3
    if sys.version_info.minor < 10:
        assert len(rte.records) == 17
    else:
        assert len(rte.records) == 16

def test_generator_grandparent_parent_child():
    def child_generator(n):
        for i in range(n):
            yield i

    def parent_generator(n):
        for value in child_generator(n):
            yield 2 * value

    def grandparent_generator(n):
        for value in parent_generator(n):
            yield 2 * value

    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
        results = []
        for v in grandparent_generator(3):
            results.append(v)
        assert results == [0, 4, 8]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert sys.version_info.major == 3
    if sys.version_info.minor < 10:
        assert len(rte.records) == 26
    else:
        assert len(rte.records) == 24

def test_generator_class_different_named_methods():
    class ChildGenerator:
        def child_generate(self, n):
            yield n
            yield n * 2
            yield n * 4
    
    class ParentGenerator:
        def parent_generate(self, n):
            child = ChildGenerator()
            for value in child.child_generate(n):
                yield 3 * value
            
    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
        results = []
        gen_class = ParentGenerator()
        for v in gen_class.parent_generate(3):
            results.append(v)
        assert results == [9, 18, 36]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert sys.version_info.major == 3
    if sys.version_info.minor < 10:
        assert len(rte.records) == 17
    else:
        assert len(rte.records) == 16

def test_generator_class_same_named_methods():
    class ChildGenerator:
        def generate(self, n):
            yield n
            yield n * 2
            yield n * 4
    
    class ParentGenerator:
        def generate(self, n):
            child = ChildGenerator()
            for value in child.generate(n):
                yield 3 * value
            
    with RecordTraceEvents(('call', 'exception', 'return')) as rte:
        # Avoid <listcomp> for the moment
        results = []
        gen_class = ParentGenerator()
        for v in gen_class.generate(3):
            results.append(v)
        assert results == [9, 18, 36]
    if TRACE_TO_STDOUT:
        print()
#         pprint.pprint(rte.records)
        print(rte)
        print('Code line numbers:', sorted(set([r.lineno_code for r in rte.records])))
    assert sys.version_info.major == 3
    if sys.version_info.minor < 10:
        assert len(rte.records) == 17
    else:
        assert len(rte.records) == 16



"""
A wrapper around the tracemalloc standard library module.
"""
import functools
import logging
import sys
import tracemalloc

import typing


class TraceMalloc:
    """A wrapper around the tracemalloc module that can compensate for tracemalloc's memory usage."""

    # Central flag to control all instances of TraceMalloc's
    TRACE_ON = True
    ALLOWED_GRANULARITY = ('filename', 'lineno', 'traceback')

    def __init__(self, statistics_granularity: str = 'lineno'):
        """statistics_granularity can be 'filename', 'lineno' or 'traceback'."""
        if statistics_granularity not in self.ALLOWED_GRANULARITY:
            raise ValueError(
                f'statistics_granularity must be in {self.ALLOWED_GRANULARITY} not {statistics_granularity}'
            )
        self.statistics_granularity = statistics_granularity
        if self.TRACE_ON:
            if not tracemalloc.is_tracing():
                tracemalloc.start()
            self.tracemalloc_snapshot_start: typing.Optional[tracemalloc.Snapshot] = None
            self.tracemalloc_snapshot_finish: typing.Optional[tracemalloc.Snapshot] = None
            self.memory_start: typing.Optional[int] = None
            self.memory_finish: typing.Optional[int] = None
            self.statistics: typing.List[tracemalloc.StatisticDiff] = []
            self._diff: typing.Optional[int] = None

    def __enter__(self):
        """Take a tracemalloc snapshot."""
        if self.TRACE_ON:
            self.tracemalloc_snapshot_start = tracemalloc.take_snapshot()
            self.memory_start = tracemalloc.get_tracemalloc_memory()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Take a tracemalloc snapshot and subtract the initial snapshot. Also note the tracemalloc memory usage."""
        if self.TRACE_ON:
            self.tracemalloc_snapshot_finish = tracemalloc.take_snapshot()
            self.memory_finish = tracemalloc.get_tracemalloc_memory()
            self.statistics = self.tracemalloc_snapshot_finish.compare_to(
                self.tracemalloc_snapshot_start, self.statistics_granularity
            )
            self._diff = None
        return False

    @property
    def tracemalloc_memory_usage(self) -> typing.Optional[int]:
        """Returns the tracemalloc memory usage between snapshots of None of no tracing."""
        if self.TRACE_ON:
            return self.memory_finish - self.memory_start

    @property
    def diff(self) -> int:
        """The net memory usage difference recorded by tracemalloc allowing for the memory usage of tracemalloc."""
        if self.TRACE_ON:
            if self._diff is None:
                self._diff = sum(s.size_diff for s in self.statistics) - self.tracemalloc_memory_usage
            return self._diff
        return -sys.maxsize - 1

    def net_statistics(self):
        """Returns the list of statistics ignoring those from the tracemalloc module itself."""
        ret = []
        for statistic in self.statistics:
            file_name = statistic.traceback[0].filename
            if file_name != tracemalloc.__file__:
                ret.append(statistic)
        return ret


def trace_malloc_log(log_level: int):
    """Decorator that logs the decorated function the use of Python memory in bytes at the desired log level.
    This can be switched to a NOP by setting TraceMalloc.TRACE_ON to False."""
    def memory_inner(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            with TraceMalloc() as tm:
                result = fn(*args, ** kwargs)
            logging.log(log_level, f'TraceMalloc memory delta: {tm.diff:,d} for "{fn.__name__}()"')
            return result
        return wrapper
    return memory_inner



"""
Logs process information, such as memory usage, to a log as JSON. Example with ('memory_info', 'cpu_times')::

    (Thread-7  ) ProcessLoggingThread JSON: {"memory_info": {"rss": 145448960, "vms": 4542902272, "pfaults": 37618, "pageins": 0}, "cpu_times": {"user": 0.28422032, "system": 0.099182912, "children_user": 0.0, "children_system": 0.0}}

There are several DoF here:

- Logging interval in seconds. Or by poke()?
- Logging level, DEBUG, INFO etc.
- Logging verbosity, for example just memory? Or everything about the process (self._process.as_dict())

"""
import argparse
import contextlib
import datetime
import json
import logging
import os
import queue
import re
import sys
import threading
import time
import typing

import psutil

from pymemtrace.util import gnuplot

logger = logging.getLogger(__file__)


#: Unique string in the log line
LOGGER_PREFIX = 'ProcessLoggingThread-JSON'
LOGGER_PREFIX_START = f'{LOGGER_PREFIX}-START'
LOGGER_PREFIX_STOP = f'{LOGGER_PREFIX}-STOP'
#: Regex for the unique string in the log line
RE_LOG_LINE = re.compile(rf'^.+?{LOGGER_PREFIX}(-START|-STOP)?\s*(.+)$')
#: Regex for timestam, matches '2019-06-07 11:57:58.390921'
DATETIME_NOW_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
assert datetime.datetime.strptime(str(datetime.datetime.now()), DATETIME_NOW_FORMAT)
#: The JSON key that is the timestamp
KEY_TIMESTAMP = 'timestamp'
#: The JSON key that is elapsed (wall clock) time in seconds. This is ``time.time() - self._process.create_time()``
KEY_ELAPSED_TIME = 'elapsed_time'
#: The JSON key that is the label
KEY_LABEL = 'label'
#: The JSON key that is the process ID
KEY_PROCESS_ID = 'pid'
#: psutil.Process().as_dict() has the following keys:
PSUTIL_PROCESS_AS_DICT_KEYS = [
    'cmdline', 'connections', 'cpu_percent', 'cpu_times', 'create_time', 'cwd', 'environ', 'exe', 'gids',
    'memory_full_info', 'memory_info', 'memory_percent', 'name', 'nice', 'num_ctx_switches', 'num_fds', 'num_threads',
    'open_files', 'pid', 'ppid', 'status', 'terminal', 'threads', 'uids', 'username'
]
#: Usage: GNUPLOT_PLT.format(name=dat_file_name)
GNUPLOT_PLT = """
set grid
set title "Memory and CPU Usage." font ",14"
set xlabel "Elapsed Time (s)"
# set mxtics 5
# set xrange [0:3000]
# set xtics
# set format x ""

#set logscale y
set ylabel "Memory Usage (Mb)"
# set yrange [0:500]
# set ytics 20
# set mytics 2
# set ytics 8,35,3

#set logscale y2
set y2label "CPU Usage (%), Page Faults (10,000/s)"
# set y2range [0:200]
set y2tics

set pointsize 1
set datafile separator whitespace#"	"
set datafile missing "NaN"

set terminal svg size 1000,700 # choose the file format
set output "{name}.svg" # choose the output device

# set key off

{labels}

#set key title "Window Length"
#  lw 2 pointsize 2

plot "{name}.dat" using 1:($2 / 1024**2) axes x1y1 title "RSS (Mb), left axis" with lines lt 1 lw 2, \\
    "{name}.dat" using 1:($3 / 10000) axes x1y2 title "Page Faults (10,000/s), right axis" with lines lt 3 lw 1, \\
    "{name}.dat" using 1:5 axes x1y2 title "Mean CPU (%), right axis" with lines lt 2 lw 1, \\
    "{name}.dat" using 1:6 axes x1y2 title "Instantaneous CPU (%), right axis" with lines lt 7 lw 1

reset
"""


def parse_timestamp(s: str) -> datetime.datetime:
    """Read a string such as '2019-06-07 11:57:58.390921' and return a datetime."""
    return datetime.datetime.strptime(s, DATETIME_NOW_FORMAT)


def extract_json(istream: typing.TextIO) -> typing.List[typing.Dict[str, typing.Any]]:
    """Reads a log file and returns the JSON as a list of dicts. Non-matching lines are ignored."""
    ret = []
    for line in istream.readlines():
        m = RE_LOG_LINE.match(line)
        if m:
            log_dict = json.loads(m.group(2))
            if KEY_TIMESTAMP in log_dict:
                log_dict[KEY_TIMESTAMP] = parse_timestamp(log_dict[KEY_TIMESTAMP])
            ret.append(log_dict)
    return ret


def extract_labels_from_json(json_data: typing.List[typing.Dict[str, typing.Any]]) \
        -> typing.List[typing.Dict[str, typing.Any]]:
    """Returns a list of dicts of JSON data where 'label' is a key'."""
    return [v for v in json_data if KEY_LABEL in v]


def extract_json_as_table(json_data: typing.List[typing.Dict[str, typing.Any]]) \
        -> typing.Tuple[
            typing.Dict[int, typing.List[typing.List[str]]],
            typing.Dict[int, float],
            typing.Dict[int, float],
            typing.Dict[int, float],
            typing.Dict[int, float],
        ]:
    """Create a table from JSON suitable for a Gnuplot ``.dat`` file.

    Returns:

        - { process_id : [rows of data, ...], ...}
        - { process_id : t min, ...}
        - { process_id : t max, ...}
        - { process_id : RSS min, ...}
        - { process_id : RSS max, ...}

    A row of data is:

        time, RSS, PageFaults, User, Mean CPU, Insantanous CPU, Timestamp, PID, Label

    """
    header = [
            f'{"#t(s)":12}',
            f'{"RSS":>12}',
            f'{"PageFaults/s":>12}',
            f'{"User":>12}',
            f'{"Mean_CPU%":>12}',
            f'{"Inst_CPU%":>12}',
            f'{"Timestamp":<26}',
            f'{"PID":>6}',
            f'{"Label"}',
        ]
    ret = {}
    prev_cpu = {}
    prev_elapsed_time = {}
    prev_page_faults = {}
    t_min = {}
    t_max = {}
    rss_min = {}
    rss_max = {}
    for record in json_data:
        if record[KEY_PROCESS_ID] not in ret:
            ret[record[KEY_PROCESS_ID]] = [header[:]]
            prev_cpu[record[KEY_PROCESS_ID]] = 0.0
            prev_elapsed_time[record[KEY_PROCESS_ID]] = 0.0
            prev_page_faults[record[KEY_PROCESS_ID]] = 0
            t_min[record[KEY_PROCESS_ID]] = sys.float_info.max
            t_max[record[KEY_PROCESS_ID]] = sys.float_info.min
            rss_min[record[KEY_PROCESS_ID]] = sys.float_info.max
            rss_max[record[KEY_PROCESS_ID]] = sys.float_info.min
        mean_cpu_user = record["cpu_times"]["user"] / record[KEY_ELAPSED_TIME]
        inst_cpu_user = (record["cpu_times"]["user"] - prev_cpu[record[KEY_PROCESS_ID]]) \
            / (record[KEY_ELAPSED_TIME] - prev_elapsed_time[record[KEY_PROCESS_ID]])
        # record["memory_info"]["pfaults"] is the cumulative total.
        inst_page_faults = (record["memory_info"]["pfaults"] - prev_page_faults[record[KEY_PROCESS_ID]]) \
            / (record[KEY_ELAPSED_TIME] - prev_elapsed_time[record[KEY_PROCESS_ID]])
        label = record[KEY_LABEL] if KEY_LABEL in record else ''
        ret[record[KEY_PROCESS_ID]].append(
            [
                f'{record[KEY_ELAPSED_TIME]:<12.1f}',
                f'{record["memory_info"]["rss"]:12d}',
                f'{inst_page_faults:12f}',
                f'{record["cpu_times"]["user"]:12.1f}',
                f'{mean_cpu_user:12.1%}',
                f'{inst_cpu_user:12.1%}',
                f'{record["timestamp"].strftime("%Y-%m-%dT%H:%M:%S.%f"):26}',
                f'{record[KEY_PROCESS_ID]:6d}',
                f'# {label}'
            ]
        )
        prev_cpu[record[KEY_PROCESS_ID]] = record["cpu_times"]["user"]
        prev_elapsed_time[record[KEY_PROCESS_ID]] = record[KEY_ELAPSED_TIME]
        prev_page_faults[record[KEY_PROCESS_ID]] = record["memory_info"]["pfaults"]
        t_min[record[KEY_PROCESS_ID]] = min(record[KEY_ELAPSED_TIME], t_min[record[KEY_PROCESS_ID]])
        t_max[record[KEY_PROCESS_ID]] = max(record[KEY_ELAPSED_TIME], t_max[record[KEY_PROCESS_ID]])
        rss_min[record[KEY_PROCESS_ID]] = min(record["memory_info"]["rss"], rss_min[record[KEY_PROCESS_ID]])
        rss_max[record[KEY_PROCESS_ID]] = max(record["memory_info"]["rss"], rss_max[record[KEY_PROCESS_ID]])
    return ret, t_min, t_max, rss_min, rss_max


def invoke_gnuplot(log_path: str, gnuplot_dir: str) -> int:
    """Reads a log file, extracts the data, writes it out to gnuplot_dir and invokes gnuplot on it."""
    os.makedirs(gnuplot_dir, exist_ok=True)
    ret = gnuplot.write_test_file(gnuplot_dir, 'svg')
    if ret:
        logger.error(f'Can not write gnuplot test file with error code {ret}')
        return ret
    with open(log_path) as instream:
        json_data = extract_json(instream)
    table, _t_min, _t_max, rss_min, rss_max = extract_json_as_table(json_data)
    for pid in table:
        log_name = f'{os.path.basename(log_path)}_{pid}'
        labels = extract_labels_from_json(json_data)
        label_lines = []
        y_value = (0.5 * (rss_max[pid] - rss_min[pid])) / 1024**2
        for label_dict in labels:
            t_value = label_dict[KEY_ELAPSED_TIME]
            label_lines.append(f'set arrow from {t_value},{y_value} to {t_value},0 lt -1 lw 1')
            label_lines.append(
                f'set label "{label_dict[KEY_LABEL]}" at {t_value},{y_value * 1.025}'
                f' left font ",10" rotate by 90 noenhanced front'
            )
        ret = gnuplot.invoke_gnuplot(
            gnuplot_dir, log_name, table[pid], GNUPLOT_PLT.format(name=log_name, labels='\n'.join(label_lines))
        )
        if ret:
            break
    return ret


# Message passing
process_queue = queue.Queue()


def add_message_to_queue(msg: str) -> None:
    """Adds a message onto the queue."""
    process_queue.put(msg)


class ProcessLoggingThread(threading.Thread):
    """Thread that regularly logs out process parameters."""
    def __init__(self, group=None, target=None, name=None, daemon=None,
                 interval=1.0, log_level=logging.INFO, pid=-1,
                 ):
        """Constructor.
        args[0], or interval=... must be the reporting interval in seconds, default 1.0.
        args[1], or log_level=... must be the log level to report with, default logging.INFO.
        """
        if name is None:
            name = 'ProcMon'
        super().__init__(group=group, target=target, name=name, daemon=daemon)
        self._interval = interval
        self._log_level = log_level
        self._pid = pid
        if self._pid != -1:
            self._process = psutil.Process(self._pid)
        else:
            self._process = psutil.Process()
            self._pid = self._process.pid
        self._run = True

    def _get_process_data(self, **kwargs):
        ret = {
            KEY_TIMESTAMP: datetime.datetime.now().strftime(DATETIME_NOW_FORMAT),
        }
        ret.update(
            {
                k: getattr(self._process, k)()._asdict() for k in ('memory_info', 'cpu_times')
            }
        )
        ret[KEY_ELAPSED_TIME] = time.time() - self._process.create_time()
        ret[KEY_PROCESS_ID] = self._process.pid
        # WARNING: This is super verbose and leaks information such as user, environment etc. into the log file.
        # ret.update(self._process.as_dict())
        # kwargs trump everything
        ret.update(kwargs)
        return ret

    def _write_to_log(self, prefix: str) -> None:
        """Write process data to log flushing message queue if necessary."""
        if self._run:
            if process_queue.empty():
                logger.log(self._log_level, f'{prefix} {json.dumps(self._get_process_data())}')
            else:
                while not process_queue.empty():
                    msg = process_queue.get()
                    logger.log(self._log_level, f'{prefix} {json.dumps(self._get_process_data(label=msg))}')

    def run(self) -> None:
        """thread.run(). Write to log then sleep."""
        self._write_to_log(LOGGER_PREFIX_START)
        while self._run:
            time.sleep(self._interval)
            self._write_to_log(LOGGER_PREFIX)

    def join(self, *args, **kwargs):
        """thread.join(). Write to log last time."""
        self._write_to_log(LOGGER_PREFIX_STOP)
        self._run = False
        super().join(*args, **kwargs)


@contextlib.contextmanager
def log_process(*args, **kwargs):
    """Context manager to log process data at regular intervals."""
    process_thread = ProcessLoggingThread(*args, **kwargs)
    process_thread.start()
    try:
        yield
    finally:
        process_thread.join()


def add_process_logger_to_argument_parser(parser: argparse.ArgumentParser) -> None:
    """Add a ``--log-process`` option to the argument parser."""
    parser.add_argument(
        '--log-process', default=0.0, type=float,
        help='Writes process data such as memory usage as a log INFO line every LOG_PROCESS seconds.'
             ' If 0.0 no process data is logged. [default: %(default)s]',
    )


def main() -> int:
    """Main CLI entry point. For testing."""
    parser = argparse.ArgumentParser(
        prog='process.py',
        description="""Reads an annotated log of a process and writes a Gnuplot graph.""",
    )
    parser.add_argument('-i', '--interval', type=float, help='Logging interval in seconds [default: %(default)s]',
                        default=1.0)
    parser.add_argument('-p', '--pid', type=int, help='PID to monitor, -1 it this [default: %(default)s]',
                        default=-1)
    parser.add_argument("-l", "--log_level", type=int, dest="log_level", default=20,
                        help="Log Level (debug=10, info=20, warning=30, error=40, critical=50)"
                             " [default: %(default)s]"
                        )
    parser.add_argument('path_in', type=str, help='Input path.', nargs='?')
    parser.add_argument('path_out', type=str, help='Output path.', nargs='?')
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s - %(filename)s#%(lineno)d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
    )
    args = parser.parse_args()
    if args.path_in and args.path_out:
        invoke_gnuplot(args.path_in, args.path_out)
    else:
        # Log another process or self
        logger.info('Demonstration of logging a process')
        with log_process(interval=args.interval, log_level=args.log_level, pid=args.pid):
            try:
                while True:
                    time.sleep(1000)
                    add_message_to_queue(f'Parent process wakes...')
            except KeyboardInterrupt:
                print('KeyboardInterrupt!')
                pass
    print('Bye, bye!')
    return 0


if __name__ == '__main__':
    sys.exit(main())




import sys
import time

def make_list_strings(n):
    lst = []
    for _i in range(n):
        lst.append(' ' * 1024)
    time.sleep(0.25)
    return lst

def trim_list(lst, length):
    while len(lst) > length:
        lst.pop()
    time.sleep(0.15)
    return lst

def just_sleep(t):
    time.sleep(t)

def main():
    for _i in range(3):
        l = make_list_strings(1024 * 10)
        just_sleep(0.5)
        trim_list(l, 128)
        just_sleep(0.2)
    return 0

if __name__ == '__main__':
    sys.exit(main())



import pprint

import pytest

from pymemtrace import debug_malloc_stats


# TODO: This fails with ``pytest tests/`` with the error:
# tests/test_debug_malloc_stats.py Fobject address  : 0x106649b40
# object refcount : 4
# object type     : 0x10615a248
# object type name: ValueError
# object repr     : ValueError('I/O operation on closed file.')
# lost sys.stderr
#
# But succeeds with ``pytest -s tests/``
#
# def test_get_debugmallocstats():
#     result = debug_malloc_stats.get_debugmallocstats()
#     assert isinstance(result, bytes)


@pytest.mark.parametrize(
    'args, expected',
    (
            ((0, 16, 4, 777, 235), 777 * 16),
    ),
)
def test_class_debugmallocstats_allocated_bytes(args, expected):
    dms = debug_malloc_stats.DebugMallocStat(*args)
    assert dms.allocated_bytes == expected


@pytest.mark.parametrize(
    'args, expected',
    (
            ((0, 16, 4, 777, 235), 235 * 16),
    ),
)
def test_class_debugmallocstats_available_bytes(args, expected):
    dms = debug_malloc_stats.DebugMallocStat(*args)
    assert dms.available_bytes == expected


@pytest.mark.parametrize(
    'args, expected',
    (
            ((0, 16, 4, 777, 235), 4 * debug_malloc_stats.POOL_OVERHEAD),
    ),
)
def test_class_debugmallocstats_pool_header_bytes(args, expected):
    dms = debug_malloc_stats.DebugMallocStat(*args)
    assert dms.pool_header_bytes == expected


@pytest.mark.parametrize(
    'args, expected',
    (
            ((0, 16, 4, 777, 235), 0),
    ),
)
def test_class_debugmallocstats_quantization(args, expected):
    dms = debug_malloc_stats.DebugMallocStat(*args)
    assert dms.quantization == expected


@pytest.mark.parametrize(
    'args, expected',
    (
            ((0, 16, 2, 297, 209), '    0     16           2             297           209'),
    ),
)
def test_class_debugmallocstats_repr(args, expected):
    dms = debug_malloc_stats.DebugMallocStat(*args)
    assert repr(dms) == expected


@pytest.mark.parametrize(
    'a, b, expected',
    (
            (
                    (0, 16, 2, 297, 209),
                    (0, 16, 3, 297, 209),
                    '    0     16          +1              +0            +0'
            ),
            (
                    (0, 16, 2, 297, 209),
                    (0, 16, 2, 297, 209),
                    '    0     16          +0              +0            +0'
            ),
            (
                    (0, 16, 2, 297, 209),
                    (0, 16, 2, 298, 209),
                    '    0     16          +0              +1            +0'
            ),
            (
                    (0, 16, 2, 297, 209),
                    (0, 16, 2, 297, 208),
                    '    0     16          +0              +0            -1'
            ),
    ),
)
def test_diff_debug_malloc_stat(a, b, expected):
    dms_a = debug_malloc_stats.DebugMallocStat(*a)
    dms_b = debug_malloc_stats.DebugMallocStat(*b)
    assert debug_malloc_stats.diff_debug_malloc_stat(dms_a, dms_b) == expected


@pytest.mark.parametrize(
    'a, b, expected',
    (
            (
                    (0, 16, 2, 297, 209),
                    (1, 16, 2, 297, 209),
                    'a.block_class != b.block_class: 0 != 1'
            ),
            (
                    (0, 16, 2, 297, 209),
                    (0, 32, 2, 297, 209),
                    'a.size != b.size: 16 != 32'
            ),
    ),
)
def test_diff_debug_malloc_stat_raises(a, b, expected):
    dms_a = debug_malloc_stats.DebugMallocStat(*a)
    dms_b = debug_malloc_stats.DebugMallocStat(*b)
    with pytest.raises(ValueError) as err:
        debug_malloc_stats.diff_debug_malloc_stat(dms_a, dms_b)
    assert err.value.args[0] == expected


@pytest.mark.parametrize(
    'line, expected',
    (
        (
            b'    0     16           2             294           212',
            (b'0', b'16', b'2', b'294', b'212'  ),
        ),
    ),
)
def test_re_debug_malloc_stats_line(line, expected):
    m = debug_malloc_stats.RE_DEBUG_MALLOC_STATS_LINE.match(line)
    assert m is not None
    assert m.groups() == expected


@pytest.mark.parametrize(
    'line, expected',
    (
        (b'# arenas allocated total           =                   33', 33),
        (b'# arenas allocated total           =                2,033', 2033),
    ),
)
def test_last_value_as_int(line, expected):
    result = debug_malloc_stats.last_value_as_int(line)
    assert result == expected


@pytest.mark.parametrize(
    'line, expected',
    (
        (
            b'18 arenas * 262144 bytes/arena     =            4,718,592',
            (b'18', b'262144', b'4,718,592'),
        ),
    ),
)
def test_re_debug_malloc_arenas_summary_line(line, expected):
    m = debug_malloc_stats.RE_DEBUG_MALLOC_ARENAS_SUMMARY_LINE.match(line)
    assert m is not None
    assert m.groups() == expected


DEBUG_MALLOC_ARENAS_BYTES = b"""# arenas allocated total           =                2,033
# arenas reclaimed                 =                2,015
# arenas highwater mark            =                   20
# arenas allocated current         =                   18
18 arenas * 262144 bytes/arena     =            4,718,592"""


def test_class_debugmallocarenas_ctor():
    dma = debug_malloc_stats.DebugMallocArenas(DEBUG_MALLOC_ARENAS_BYTES)
    assert dma.ntimes_arena_allocated == 2033
    assert dma.narenas_highwater == 20
    assert dma.narenas == 18
    assert dma.arena_size == 262144
    assert dma.arenas_reclaimed == 2033 - 18
    assert dma.arenas_total == 4718592


def test_class_debugmallocarenas_repr():
    dma = debug_malloc_stats.DebugMallocArenas(DEBUG_MALLOC_ARENAS_BYTES)
    assert repr(dma).encode('ascii') == DEBUG_MALLOC_ARENAS_BYTES


@pytest.mark.parametrize(
    'line, expected',
    (
        (
            b'63 unused pools * 4096 bytes       =              258,048',
            (b'63', b'4096', b'258,048'),
        ),
    ),
)
def test_re_debug_malloc_pools_summary_line(line, expected):
    m = debug_malloc_stats.RE_DEBUG_MALLOC_POOLS_SUMMARY_LINE.match(line)
    assert m is not None
    assert m.groups() == expected


DEBUG_MALLOC_POOLS_BLOCKS_BYTES = b"""# bytes in allocated blocks        =            4,195,888
# bytes in available blocks        =               75,136
19 unused pools * 4096 bytes       =               77,824
# bytes lost to pool headers       =               51,312
# bytes lost to quantization       =               56,288
# bytes lost to arena alignment    =                    0
Total                              =            4,456,448"""


def test_class_debug_malloc_pools_blocks_ctor():
    dmpb = debug_malloc_stats.DebugMallocPoolsBlocks(DEBUG_MALLOC_POOLS_BLOCKS_BYTES)
    assert dmpb is not None


@pytest.mark.parametrize(
    'attr, expected',
    (
        ('allocated_bytes', 4_195_888),
        ('available_bytes', 75_136),
        ('numfreepools', 19),
        ('pool_size', 4096),
        ('unused_pool_total', 77_824),
        ('pool_header_bytes', 51_312),
        ('quantization', 56_288),
        ('arena_alignment', 0),
        ('unused_pool_total', 19 * 4096),
        ('total', 4_456_448),
    ),
)
def test_class_debug_malloc_pools_blocks_attributes(attr, expected):
    dmpb = debug_malloc_stats.DebugMallocPoolsBlocks(DEBUG_MALLOC_POOLS_BLOCKS_BYTES)
    assert hasattr(dmpb, attr)
    assert getattr(dmpb, attr) == expected


def test_class_debug_malloc_pools_blocks_repr():
    dmpb = debug_malloc_stats.DebugMallocPoolsBlocks(DEBUG_MALLOC_POOLS_BLOCKS_BYTES)
    assert repr(dmpb).encode('ascii') == DEBUG_MALLOC_POOLS_BLOCKS_BYTES


@pytest.mark.parametrize(
    'args, expected',
    (
        (
            (9, 'PyDictObjects', 48, 432),
            '            9 free PyDictObjects * 48 bytes each =                  432',
        ),
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            '  52 free 2-sized PyTupleObjects * 40 bytes each =                2,080',
        ),
    ),
)
def test_class_debug_type_stat_repr(args, expected):
    dts = debug_malloc_stats.DebugTypeStat(*args)
    assert repr(dts) == expected


@pytest.mark.parametrize(
    'line, expected',
    (
        (
            b'   34 free 2-sized PyTupleObjects * 40 bytes each =                1,360',
            (b'34', b'2-sized PyTupleObjects', b'40', b'1,360'),
        ),
    ),
)
def test_re_debug_malloc_type_line(line, expected):
    m = debug_malloc_stats.RE_DEBUG_MALLOC_TYPE_LINE.match(line)
    assert m is not None
    assert m.groups() == expected


@pytest.mark.parametrize(
    'a, b, expected',
    (
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            (52, '2-sized PyTupleObjects', 40, 2080),
            '  +0 free 2-sized PyTupleObjects * 40 bytes each =                   +0'
        ),
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            (51, '2-sized PyTupleObjects', 40, 2080),
            '  -1 free 2-sized PyTupleObjects * 40 bytes each =                   +0'
        ),
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            (52, '2-sized PyTupleObjects', 40, 2090),
            '  +0 free 2-sized PyTupleObjects * 40 bytes each =                  +10'
        ),
    ),
)
def test_diff_diff_debug_type_stat(a, b, expected):
    dts_a = debug_malloc_stats.DebugTypeStat(*a)
    dts_b = debug_malloc_stats.DebugTypeStat(*b)
    assert debug_malloc_stats.diff_debug_type_stat(dts_a, dts_b) == expected


@pytest.mark.parametrize(
    'a, b, expected',
    (
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            (52, '3-sized PyTupleObjects', 40, 2080),
            'a.object_type != b.object_type: 2-sized PyTupleObjects != 3-sized PyTupleObjects'
        ),
        (
            (52, '2-sized PyTupleObjects', 40, 2080),
            (52, '2-sized PyTupleObjects', 64, 2080),
            'a.bytes_each != b.bytes_each: 40 != 64'
        ),
    ),
)
def test_diff_debug_malloc_stat_raises(a, b, expected):
    dts_a = debug_malloc_stats.DebugTypeStat(*a)
    dts_b = debug_malloc_stats.DebugTypeStat(*b)
    with pytest.raises(ValueError) as err:
        debug_malloc_stats.diff_debug_type_stat(dts_a, dts_b)
    assert err.value.args[0] == expected


@pytest.mark.parametrize(
    'line, expected',
    (
        (
            b'Small block threshold = 512, in 32 size classes.',
            (b'512', b'32',),
        ),
    ),
)
def test_re_debug_malloc_header_line(line, expected):
    m = debug_malloc_stats.RE_DEBUG_MALLOC_HEADER_LINE.match(line)
    assert m is not None
    assert m.groups() == expected


SYS_DEBUGMALLOCSTATS_EXAMPLE = b"""Small block threshold = 512, in 32 size classes.

class   size   num pools   blocks in use  avail blocks
-----   ----   ---------   -------------  ------------
    0     16           2             297           209
    1     32           9            1131             3
    2     48          58            4788            84
    3     64         303           19065            24
    4     80         214           10672            28
    5     96          36            1507             5
    6    112          19             659            25
    7    128          16             478            18
    8    144          70            1944            16
    9    160          12             292             8
   10    176         116            2665             3
   11    192           6             123             3
   12    208          30             555            15
   13    224          55             982             8
   14    240          10             149            11
   15    256           8             116             4
   16    272           4              45            11
   17    288           4              49             7
   18    304          25             321             4
   19    320           3              29             7
   20    336           3              26            10
   21    352           2              21             1
   22    368           3              30             3
   23    384           3              24             6
   24    400           4              35             5
   25    416           9              67            14
   26    432          10              78            12
   27    448           8              69             3
   28    464           8              57             7
   29    480           6              43             5
   30    496           7              53             3
   31    512          33             228             3

# arenas allocated total           =                2,034
# arenas reclaimed                 =                2,016
# arenas highwater mark            =                   18
# arenas allocated current         =                   18
18 arenas * 262144 bytes/arena     =            4,718,592

# bytes in allocated blocks        =            4,310,720
# bytes in available blocks        =               68,704
56 unused pools * 4096 bytes       =              229,376
# bytes lost to pool headers       =               52,608
# bytes lost to quantization       =               57,184
# bytes lost to arena alignment    =                    0
Total                              =            4,718,592

       4 free PyCFunctionObjects * 56 bytes each =                  224
            3 free PyDictObjects * 48 bytes each =                  144
           5 free PyFloatObjects * 24 bytes each =                  120
          0 free PyFrameObjects * 368 bytes each =                    0
           23 free PyListObjects * 40 bytes each =                  920
          7 free PyMethodObjects * 48 bytes each =                  336
   3 free 1-sized PyTupleObjects * 32 bytes each =                   96
  54 free 2-sized PyTupleObjects * 40 bytes each =                2,160
   5 free 3-sized PyTupleObjects * 48 bytes each =                  240
   5 free 4-sized PyTupleObjects * 56 bytes each =                  280
   3 free 5-sized PyTupleObjects * 64 bytes each =                  192
   4 free 6-sized PyTupleObjects * 72 bytes each =                  288
  11 free 7-sized PyTupleObjects * 80 bytes each =                  880
  10 free 8-sized PyTupleObjects * 88 bytes each =                  880
   4 free 9-sized PyTupleObjects * 96 bytes each =                  384
 0 free 10-sized PyTupleObjects * 104 bytes each =                    0
 1 free 11-sized PyTupleObjects * 112 bytes each =                  112
 2 free 12-sized PyTupleObjects * 120 bytes each =                  240
 0 free 13-sized PyTupleObjects * 128 bytes each =                    0
 1 free 14-sized PyTupleObjects * 136 bytes each =                  136
 0 free 15-sized PyTupleObjects * 144 bytes each =                    0
 1 free 16-sized PyTupleObjects * 152 bytes each =                  152
 2 free 17-sized PyTupleObjects * 160 bytes each =                  320
 1 free 18-sized PyTupleObjects * 168 bytes each =                  168
 2 free 19-sized PyTupleObjects * 176 bytes each =                  352
"""

def test_class_sysdebugmallocstats_example():
    sdms = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert sdms is not None


def test_class_sysdebugmallocstats_example_repr():
    sdms = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert bytes(repr(sdms).encode('ascii')) == SYS_DEBUGMALLOCSTATS_EXAMPLE


def test_class_sysdebugmallocstats_malloc_stats_block_class():
    sdms = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert [v.block_class for v in sdms.malloc_stats] == list(range(32))


def test_class_sysdebugmallocstats_malloc_stats_size():
    sdms = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert [v.size for v in sdms.malloc_stats] == list(range(16, 32 * 16 + 16, 16))


def test_class_sysdebugmallocstats_object_types():
    sdms = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert sorted(sdms.object_types()) == [b'1-sized PyTupleObjects', b'10-sized PyTupleObjects',
                                           b'11-sized PyTupleObjects', b'12-sized PyTupleObjects',
                                           b'13-sized PyTupleObjects', b'14-sized PyTupleObjects',
                                           b'15-sized PyTupleObjects', b'16-sized PyTupleObjects',
                                           b'17-sized PyTupleObjects', b'18-sized PyTupleObjects',
                                           b'19-sized PyTupleObjects', b'2-sized PyTupleObjects',
                                           b'3-sized PyTupleObjects', b'4-sized PyTupleObjects',
                                           b'5-sized PyTupleObjects', b'6-sized PyTupleObjects',
                                           b'7-sized PyTupleObjects', b'8-sized PyTupleObjects',
                                           b'9-sized PyTupleObjects', b'PyCFunctionObjects', b'PyDictObjects',
                                           b'PyFloatObjects', b'PyFrameObjects', b'PyListObjects', b'PyMethodObjects']


def test_diff_debugmallocstats_same():
    a = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    b = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    assert debug_malloc_stats.diff_debugmallocstats(a, b) == []


SYS_DEBUGMALLOCSTATS_EXAMPLE_DIFF = b"""Small block threshold = 512, in 32 size classes.

class   size   num pools   blocks in use  avail blocks
-----   ----   ---------   -------------  ------------
    0     16           3             297           209
    1     32           9            1131             3
    2     48          58            4788            84
    3     64         303           19065            24
    4     80         214           10672            28
    5     96          36            1507             5
    6    112          19             659            25
    7    128          16             478            18
    8    144          70            1944            16
    9    160          12             292             8
   10    176         116            2665             3
   11    192           6             123             3
   12    208          30             555            15
   13    224          55             982             8
   14    240          10             149            11
   15    256           8             116             4
   16    272           4              45            11
   17    288           4              49             7
   18    304          25             321             4
   19    320           3              29             7
   20    336           3              26            10
   21    352           2              21             1
   22    368           3              30             3
   23    384           3              24             6
   24    400           4              35             5
   25    416           9              67            14
   26    432          10              78            12
   27    448           8              69             3
   28    464           8              57             7
   29    480           6              43             5
   30    496           7              53             3
   31    512          33             228             3

# arenas allocated total           =                2,034
# arenas reclaimed                 =                2,016
# arenas highwater mark            =                   19
# arenas allocated current         =                   19
19 arenas * 262144 bytes/arena     =            4,718,592

# bytes in allocated blocks        =            4,310,720
# bytes in available blocks        =               68,704
55 unused pools * 4096 bytes       =              229,376
# bytes lost to pool headers       =               52,608
# bytes lost to quantization       =               57,184
# bytes lost to arena alignment    =                    0
Total                              =            4,718,592

       4 free PyCFunctionObjects * 56 bytes each =                  224
            3 free PyDictObjects * 48 bytes each =                  144
           5 free PyFloatObjects * 24 bytes each =                  120
          0 free PyFrameObjects * 368 bytes each =                    0
           23 free PyListObjects * 40 bytes each =                  920
          7 free PyMethodObjects * 48 bytes each =                  336
   3 free 1-sized PyTupleObjects * 32 bytes each =                   96
  54 free 2-sized PyTupleObjects * 40 bytes each =                2,160
   5 free 3-sized PyTupleObjects * 48 bytes each =                  240
   5 free 4-sized PyTupleObjects * 56 bytes each =                  280
   3 free 5-sized PyTupleObjects * 64 bytes each =                  192
   4 free 6-sized PyTupleObjects * 72 bytes each =                  288
  11 free 7-sized PyTupleObjects * 80 bytes each =                  880
  10 free 8-sized PyTupleObjects * 88 bytes each =                  880
   4 free 9-sized PyTupleObjects * 96 bytes each =                  384
 0 free 10-sized PyTupleObjects * 104 bytes each =                    0
 1 free 11-sized PyTupleObjects * 112 bytes each =                  112
 2 free 12-sized PyTupleObjects * 120 bytes each =                  240
 0 free 13-sized PyTupleObjects * 128 bytes each =                    0
 1 free 14-sized PyTupleObjects * 136 bytes each =                  136
 0 free 15-sized PyTupleObjects * 144 bytes each =                    0
 1 free 16-sized PyTupleObjects * 152 bytes each =                  152
 2 free 17-sized PyTupleObjects * 160 bytes each =                  320
 1 free 18-sized PyTupleObjects * 168 bytes each =                  168
 3 free 19-sized PyTupleObjects * 176 bytes each =                  352
"""

def test_diff_debugmallocstats_diff():
    a = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE)
    b = debug_malloc_stats.SysDebugMallocStats(SYS_DEBUGMALLOCSTATS_EXAMPLE_DIFF)
    result = debug_malloc_stats.diff_debugmallocstats(a, b)
    expected = [
        'class   size   num pools   blocks in use  avail blocks',
        '-----   ----   ---------   -------------  ------------',
        '    0     16          +1              +0            +0',
        '',
        '# arenas reclaimed                 =                   -1',
        '# arenas highwater mark            =                   +1',
        '# arenas allocated current         =                   +1',
        '+1 arenas * 262144 bytes/arena     =             +262,144',
        '',
        '-1 unused pools * 4096 bytes       =               -4,096',
        'Total                              =               -4,096',
        '',
        '+1 free 19-sized PyTupleObjects * 176 bytes each =                   +0',
    ]
    assert result == expected



import argparse
import logging
import os
import sys
import time

from pymemtrace import cMemLeak


logger = logging.getLogger(__file__)


def create_c_buffer_and_del(size: int):
    input(f'Waiting to create buffer PID={os.getpid()}, size={size} <cr> to continue... ')
    b = cMemLeak.CMalloc(size)
    input(f'Waiting to del buffer PID={os.getpid()}, size={size} <cr> to continue... ')
    del b
    input(f'DONE del buffer PID={os.getpid()}, size={size} <cr> to continue... ')


def create_c_buffer(size: int):
    b = cMemLeak.CMalloc(size)
    # logger.info(f'create_c_buffer() {b.size} at 0x{b.buffer:x}')
    print(f'create_c_buffer() {b.size} at 0x{b.buffer:x}')
    return b


def exercise_c_memory():
    logger.info('exercise_c_memory()')
    str_list = []
    for i in range(8):
        str_list.append(create_c_buffer(1652))
        # time.sleep(0.5)
        input(f'Waiting to pop PID={os.getpid()}, <cr> to continue... ')
        str_list.pop()
        input(f'Pop\'d PID={os.getpid()}, <cr> to continue... ')
    while len(str_list):
        str_list.pop()
    logger.info('DONE: exercise_c_ memory()')


def create_string(size: int) -> str:
    return ' ' * size


def exercise_memory():
    logger.info('exercise_memory()')
    str_list = []
    for i in range(8):
        str_list.append(create_string(1024**2))
        time.sleep(0.5)
    while len(str_list):
        str_list.pop()
    logger.info('DONE: exercise_memory()')


def main():
    parser = argparse.ArgumentParser(
        description='Excercise Python and C memory',
        # formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    # parser.add_argument("-s", "--subdir", type=str, dest="subdir",
    #                     default=SUB_DIR_FOR_COMMON_FILES,
    #                     help="Sub-directory for writing the common files. [default: %(default)s]")
    parser.add_argument("-p", "--pause", action="store_true", dest="pause",
                        default=False,
                        help="Pause before starting. [default: %(default)s]")
    # parser.add_argument("-v", "--verbose", action="store_true", dest="verbose",
    #                     default=False,
    #                     help="Verbose, lists duplicate files and sizes. [default: %(default)s]")
    parser.add_argument(
        "-l", "--log_level",
        type=int,
        dest="log_level",
        default=20,
        help="Log Level (debug=10, info=20, warning=30, error=40, critical=50)"
             " [default: %(default)s]"
    )
    # parser.add_argument(
    #     dest="path",
    #     nargs=1,
    #     help="Path to source directory. WARNING: This will be rewritten in-place."
    # )
    args = parser.parse_args()
    if args.pause:
        input(f'Waiting to continue PID={os.getpid()}, <cr> to continue... ')

    clock_start = time.perf_counter()
    # Initialise logging etc.
    logging.basicConfig(level=args.log_level,
                        format='%(asctime)s - %(filename)-16s - %(lineno)4d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
                        # datefmt='%y-%m-%d % %H:%M:%S',
                        stream=sys.stdout)
    try:
        while True:
            # exercise_memory()
            exercise_c_memory()
            # create_c_buffer_and_del(1653)
    except KeyboardInterrupt:
        print('Interrupted!')
    print(f'Runtime: {time.perf_counter() - clock_start:.3f} (s)')
    print('Bye, bye!')
    return 0


if __name__ == '__main__':
    sys.exit(main())


import sys

import psutil

from pymemtrace import custom
from pymemtrace import cPyMemTrace


def create_string(l: int) -> str:
    return ' ' * l


COUNT = 16


def test_under_512():
    print(f'test_under_512 {COUNT}')
    l = []
    for i in range(COUNT):
        l.append(create_string(256))
    while len(l):
        l.pop()


def test_over_512():
    print(f'test_over_512 {COUNT}')
    l = []
    for i in range(COUNT):
        l.append(create_string(1024))
    while len(l):
        l.pop()


def test_well_over_512():
    print(f'test_well_over_512 {COUNT}')
    l = []
    for i in range(COUNT):
        l.append(create_string(1024**2))
    while len(l):
        l.pop()


def f(l):
    print('Hi')
    s = ' ' * l
    print('Bye')


def g():
    print(f'Creating Custom.')
    # pid = psutil.Process()
    # print(f'Creating Custom: {pid.memory_info()}')
    obj = custom.Custom('First', 'Last')
    print(obj.name())
    # print(f'Done: {pid.memory_info()}')
    print(f'Done.')


def example_for_documentation():
    print(f'example_for_documentation()')
    with cPyMemTrace.Profile(0):
        l = []
        for i in range(8):
            l.append(create_string(1024**2))
        while len(l):
            l.pop()


def main():
    # cPyMemTrace._attach_profile()
    # f(1024**2)
    # # f(1024**2)
    # g()
    # cPyMemTrace._detach_profile()

    # with cPyMemTrace.Profile():
    #     # f(1024**2)
    #     # f(1024**2)
    #     # g()
    #     test_under_512()
    #     test_over_512()
    #     test_well_over_512()

    example_for_documentation()

    return 0


if __name__ == '__main__':
    sys.exit(main())




"""
Example code to exercise DTrace.
"""
import logging
import os
import pprint
import sys
import sysconfig
import time

from pymemtrace import cMemLeak

logger = logging.getLogger(__file__)


def create_c_buffer_and_del(size: int):
    input(f'Waiting to create buffer PID={os.getpid()}, size={size} <cr> to continue... ')
    b = cMemLeak.CMalloc(size)
    input(f'Waiting to del buffer PID={os.getpid()}, size={size} <cr> to continue... ')
    del b
    input(f'DONE del buffer PID={os.getpid()}, size={size} <cr> to continue... ')


def create_c_buffer(size: int):
    b = cMemLeak.CMalloc(size)
    # logger.info(f'create_c_buffer() {b.size} at 0x{b.buffer:x}')
    print(f'create_c_buffer() {b.size} at 0x{b.buffer:x}')
    return b


def exercise_c_memory():
    logger.info('exercise_c_memory()')
    str_list = []
    for i in range(8):
        str_list.append(create_c_buffer(1652))
        # time.sleep(0.5)
        input(f'Waiting to pop PID={os.getpid()}, <cr> to continue... ')
        str_list.pop()
        input(f'Pop\'d PID={os.getpid()}, <cr> to continue... ')
    while len(str_list):
        str_list.pop()
    logger.info('DONE: exercise_c_ memory()')


def create_cmalloc_list():
    l = []
    for i in range(4):
        block = cMemLeak.CMalloc(1477)
        print(f'Created CMalloc size={block.size:d} buffer=0x{block.buffer:x}')
        l.append(block)
    while len(l):
        # Remove in reverse order
        block = l.pop(0)
        print(f'Pop\'d CMalloc size={block.size:d} buffer=0x{block.buffer:x}')
    l.clear()


def create_pyrawmalloc_list():
    l = []
    for i in range(4):
        block = cMemLeak.PyRawMalloc(128)
        print(f'Created PyRawMalloc size={block.size:d} buffer=0x{block.buffer:x}')
        l.append(block)
    while len(l):
        # Remove in reverse order
        block = l.pop(0)
        print(f'Pop\'d PyRawMalloc size={block.size:d} buffer=0x{block.buffer:x}')
    l.clear()


def create_pymalloc_list():
    print(f'Python at {sys.executable} is configured with CONFIG_ARGS: {sysconfig.get_config_var("CONFIG_ARGS")}')
    l = []
    for i in range(4):
        block = cMemLeak.PyMalloc(371)
        print(f'Created PyMalloc size={block.size:d} buffer=0x{block.buffer:x}')
        l.append(block)
    while len(l):
        # Remove in reverse order
        block = l.pop(0)
        print(f'Pop\'d PyMalloc size={block.size:d} buffer=0x{block.buffer:x}')
    l.clear()


def create_py_array_list(size: int):
    l = []
    for i in range(4):
        block = b' ' * size
        print(f'Created {type(block)} size={len(block):d} buffer=0x{id(block):x}')
        l.append(block)
    while len(l):
        # Remove in reverse order
        block = l.pop(0)
        print(f'Pop\'d {type(block)} size={len(block):d} buffer=0x{id(block):x}')
    l.clear()


def main():
    with_dtrace = sysconfig.get_config_var('WITH_DTRACE')
    if with_dtrace is None or with_dtrace != 1:
        raise RuntimeError(f'Python at {sys.executable} must be build with DTrace support.')
    logging.basicConfig(
        level=20,
        format='%(asctime)s - %(filename)-16s - %(lineno)4d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
        stream=sys.stdout,
    )
    logger.info('Python at %s is configured with CONFIG_ARGS: %s', sys.executable, sysconfig.get_config_var('CONFIG_ARGS'))
    # Wait after logging initialised
    input(f'Waiting to start tracing PID: {os.getpid()} (<cr> to continue):')
    # exercise_c_memory()

    # create_cmalloc_list()
    # create_pyrawmalloc_list()
    # create_pymalloc_list()
    create_py_array_list(27)
    return 0


if __name__ == '__main__':
    sys.exit(main())





import pytest

from pymemtrace import trace_malloc


def test_trace_malloc_simple():
    list_of_strings = []
    # with trace_malloc.TraceMalloc('filename') as tm:
    with trace_malloc.TraceMalloc('lineno') as tm:
        list_of_strings.append(' ' * 1024)
    # print()
    # print(f'tm.memory_start={tm.memory_start}')
    # print(f'tm.memory_finish={tm.memory_finish}')
    # print(f'tm.net_statistics(): {tm.net_statistics()}')
    # for stat in tm.net_statistics():
    #     print(stat)
    # assert tm.diff == 0, f'tm.diff={tm.diff}'
    # Different versions of Python will have a different number of entries.
    assert len(tm.net_statistics()) in (3, 4, 5, 6,)
    assert len(tm.statistics) > 3


if __name__ == '__main__':
    test_trace_malloc_simple()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
import sys
from setuptools import setup, find_packages
from distutils.core import Extension

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'psutil',
]

setup_requirements = [
    'pytest',
    'pytest-runner',
]

test_requirements = [
    'pytest',
]

extra_compile_args = [
    '-Wall',
    '-Wextra',
    '-Werror',
    '-Wfatal-errors',
    '-Wpedantic',
    # Some internal Python library code does not like this with C++11.
    # '-Wno-c++11-compat-deprecated-writable-strings',
    # '-std=c++11',
    '-std=c99',
    # Until we use m_coalesce
    # '-Wno-unused-private-field',

    # # Temporary
    # '-Wno-unused-variable',
    # '-Wno-unused-parameter',
]

if sys.platform.startswith('linux'):
    extra_compile_args.extend(
        [
            # Linux, GCC complains about casting PyCFunction.
            '-Wno-cast-function-type',
        ]
    )

DEBUG = False

if DEBUG:
    extra_compile_args.extend(['-g3', '-O0', '-DDEBUG=1', '-UNDEBUG'])
else:
    extra_compile_args.extend(['-O3', '-UDEBUG', '-DNDEBUG'])

setup(
    name='pymemtrace',
    version='0.1.7',
    description="Python memory tracing.",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    author="Paul Ross",
    author_email='apaulross@gmail.com',
    url='https://github.com/paulross/pymemtrace',
    packages=find_packages(),  # include=['pymemtrace']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='pymemtrace',
    # https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: C',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        # https://devguide.python.org/versions/
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Software Development',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    # Extensions
    ext_modules=[
        Extension(
            "pymemtrace.custom",
            sources=[
              'pymemtrace/src/cpy/cCustom.c',
            ],
            include_dirs=[
                '/usr/local/include',
                os.path.join('pymemtrace', 'src', 'include'),
            ],
            library_dirs=[os.getcwd(), ],  # path to .a or .so file(s)
            extra_compile_args=extra_compile_args,
        ),
        Extension(
            "pymemtrace.cPyMemTrace",
            sources=[
              'pymemtrace/src/c/get_rss.c',
              'pymemtrace/src/c/pymemtrace_util.c',
              'pymemtrace/src/cpy/cPyMemTrace.c',
            ],
            include_dirs=[
                '/usr/local/include',
                os.path.join('pymemtrace', 'src', 'include'),
            ],
            library_dirs=[os.getcwd(), ],  # path to .a or .so file(s)
            extra_compile_args=extra_compile_args,
        ),
        Extension(
            "pymemtrace.cMemLeak",
            sources=[
              'pymemtrace/src/cpy/cMemLeak.c',
            ],
            include_dirs=[
                '/usr/local/include',
                os.path.join('pymemtrace', 'src', 'include'),
            ],
            library_dirs=[os.getcwd(), ],  # path to .a or .so file(s)
            extra_compile_args=extra_compile_args,
        ),
    ]
)



import ctypes
import io
import sys

import pytest

from pymemtrace import redirect_stdout


def test_stdout_redirector_python_bytes():
    stream = io.BytesIO()
    with redirect_stdout.stdout_redirector(stream):
        sys.stdout.write('Foo')
    result = stream.getvalue()
    assert result == b'Foo', f'Result: {result!r}'


# def test_stdout_redirector_python_text():
#     stream = io.StringIO()
#     with redirect_stdout.stdout_redirector(stream):
#         sys.stdout.write('Foo')
#     result = stream.getvalue()
#     assert result == b'Foo', f'Result: {result!r}'


def test_stdout_redirector_c_bytes():
    libc = ctypes.CDLL(None)
    stream = io.BytesIO()
    with redirect_stdout.stdout_redirector(stream):
        libc.puts(b'This comes from C')
    result = stream.getvalue()
    assert result == b'This comes from C\n', f'Result: {result!r}'


def test_stdout_redirector_python_c_bytes():
    libc = ctypes.CDLL(None)
    stream = io.BytesIO()
    with redirect_stdout.stdout_redirector(stream):
        print('foobar')
        sys.stdout.write('sys.stdout.write()\n')
        libc.puts(b'This comes from C')
    result = stream.getvalue()
    assert result == b'This comes from C\nfoobar\nsys.stdout.write()\n', f'Result: {result!r}'


def test_stdout_redirector_python_c_bytes_flush():
    libc = ctypes.CDLL(None)
    stream = io.BytesIO()
    with redirect_stdout.stdout_redirector(stream):
        print('foobar')
        sys.stdout.write('sys.stdout.write()\n')
        sys.stdout.flush()
        libc.puts(b'This comes from C')
    result = stream.getvalue()
    assert result == b'foobar\nsys.stdout.write()\nThis comes from C\n', f'Result: {result!r}'


if __name__ == '__main__':
    test_stdout_redirector_python_bytes()
    # test_stdout_redirector_python_text()
    test_stdout_redirector_c_bytes()
    test_stdout_redirector_python_c_bytes()
    test_stdout_redirector_python_c_bytes_flush()



"""
Parses DTrace output.

Example from toolkit/py_flow_malloc_free.d:
dtrace:::BEGIN
 77633  cmn_cmd_opts.py:141  -> set_log_level malloc(560) pntr 0x7fca83ef4240
 77633      __init__.py:422  -> validate malloc(1114) pntr 0x7fca858e4200
 77633      __init__.py:422  -> validate free(0x7fca858e4200)
 77633     threading.py:817  -> __init__ malloc(576) pntr 0x7fca83ef4470

"""
import logging
import pprint
import re
import sys
import typing


logger = logging.getLogger(__file__)


#: Matches " 77633  cmn_cmd_opts.py:141  -> set_log_level malloc(560) pntr 0x7fca83ef4240"
#: Six groups:
#: To ('77633', 'cmn_cmd_opts.py', '141', 'set_log_level', '560', '0x7fca83ef4240')
RE_PY_FLOW_MALLOC_FREE_MALLOC = re.compile(r'^\s+(\d+)\s+(.+?):(\d+)\s+-> (\S+) malloc\((\d+)\) pntr (.+)$')


class Malloc(typing.NamedTuple):
    log_line: int
    pid: int
    file: str
    line: int
    function: str
    size: int
    address: int

    def __str__(self):
        return f'Malloc: {self.log_line} {self.file}:{self.line} {self.function} {self.size} 0x{self.address:x}'

def match_to_malloc(line: int, m: re.match) -> Malloc:
    """Given a match that has groups: ('77633', 'cmn_cmd_opts.py', '141', 'set_log_level', '560', '0x7fca83ef4240')
    this returns a Malloc() object."""
    return Malloc(
        line,
        int(m.group(1)),
        m.group(2),
        int(m.group(3)),
        m.group(4),
        int(m.group(5)),
        int(m.group(6), 16),
    )


#: Matches " 77633      __init__.py:422  -> validate free(0x7fca858e4200)"
#: Five groups:
#: To ('77633', '__init__.py', '422', 'validate', '0x7fca858e4200')
RE_PY_FLOW_MALLOC_FREE_FREE = re.compile(r'^\s+(\d+)\s+(.+?):(\d+)\s+-> (\S+) free\((\S+)\)$')


class Free(typing.NamedTuple):
    log_line: int
    pid: int
    file: str
    line: int
    function: str
    address: int

    def __str__(self):
        return f'Free: {self.log_line} {self.file}:{self.line} {self.function}'# 0x{self.address:x}'


def match_to_free(line: int, m: re.match) -> Free:
    """Given a match that has groups: ('77633', '__init__.py', '422', 'validate', '0x7fca858e4200')
    this returns a Malloc() object."""
    return Free(
        line,
        int(m.group(1)),
        m.group(2),
        int(m.group(3)),
        m.group(4),
        int(m.group(5), 16),
    )


def parse_py_flow_malloc_free(file: typing.BinaryIO) -> typing.Dict[int, Malloc]:
    file.seek(0)
    malloc_dict: typing.Dict[int, Malloc] = {}
    for l, bin_line in enumerate(file):
        line = bin_line.decode('ascii', 'ignore')
        # print(f'TRACE: {line!r}')
        m = RE_PY_FLOW_MALLOC_FREE_MALLOC.match(line)
        if m is not None:
            malloc = match_to_malloc(l, m)
            if malloc.address in malloc_dict:
                logger.error('Line %d malloc address 0x%x already in malloc dict', l, malloc.address)
            else:
                malloc_dict[malloc.address] = malloc
        else:
            m = RE_PY_FLOW_MALLOC_FREE_FREE.match(line)
            if m is not None:
                free = match_to_free(l, m)
                if free.address == 0:
                    logger.error('Ignoring 0x0 %s', free)
                else:
                    if free.address not in malloc_dict:
                        logger.error('Line %d free address 0x%x not in malloc dict', l, free.address)
                    else:
                        logger.info('%s free\'d %s', malloc_dict[free.address], free)
                        del malloc_dict[free.address]
    return malloc_dict



def main():
    logging.basicConfig(
        level=20,
        format='%(asctime)s - %(filename)-16s - %(lineno)4d - %(process)5d - (%(threadName)-10s) - %(levelname)-8s - %(message)s',
        stream=sys.stdout,
    )
    with open(sys.argv[1], 'rb') as file:
        malloc_dict = parse_py_flow_malloc_free(file)
        pprint.pprint(malloc_dict)
    return 0


if __name__ == '__main__':
    sys.exit(main())
